import skycoin
import tests.utils


def test_TestAddUint64():
    err, n = skycoin.SKY_util_AddUint64(10, 11)
    assert err == skycoin.SKY_OK
    assert int(21) == n
    err, n = skycoin.SKY_util_AddUint64(int(0xFFFFFFFFFFFFFFFF), 1)
    assert err == skycoin.SKY_ErrUint64AddOverflow


class math_test:
    a = 0
    b = 0
    err = 0


def test_TestUint64ToInt64():
    cases = []
    values = math_test()
    cases.append(values)
    values.a = int(1)
    values.b = 1
    cases.append(values)
    values.a = int(0xFFFFFFFFFFFFFFFF)
    values.b = int(0xFFFFFFFFFFFFFFFF)
    values.err = skycoin.SKY_ErrUint64AddOverflow
    cases.append(values)
    values.a = int(0xFFFFFFFFFFFFFFFF)
    values.b = 0
    values.err = skycoin.SKY_ErrUint64AddOverflow
    cases.append(values)
    for val in cases:
        s = int(val.a)
        err, x = skycoin.SKY_util_Uint64ToInt64(s)
        if err != skycoin.SKY_OK:
            assert skycoin.SKY_ErrUint64AddOverflow == val.err
        else:
            assert val.b == x


def test_TestInt64ToUint64():
    cases = []
    values = math_test()
    cases.append(values)
    values.a = 1
    values.b = 1
    cases.append(values)
    values.a = int(-0xFFFFFFFFFFFFFFFF)
    values.b = int(0)
    values.err = skycoin.SKY_ErrUint64AddOverflow
    cases.append(values)
    values.a = int(-1)
    values.b = 0
    values.err = skycoin.SKY_ErrUint64AddOverflow
    cases.append(values)
    for val in cases:
        s = int(val.a)
        err, x = skycoin.SKY_util_Int64ToUint64(s)
        if err != skycoin.SKY_OK:
            assert skycoin.SKY_ErrUint64AddOverflow == val.err
        else:
            assert val.b == x



import skycoin
import tests.utils
import base64


def freshSumRipemd160(b):
    rp160 = skycoin.cipher_Ripemd160()
    skycoin.SKY_cipher_HashRipemd160(b, rp160)
    return rp160


def freshSumSHA256(b):
    sha256 = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(b, sha256)
    return sha256


def test_TestHashRipemd160():
    r160 = skycoin.cipher_Ripemd160()
    _, b = skycoin.SKY_cipher_RandByte(128)
    assert skycoin.SKY_cipher_HashRipemd160(b, r160) == skycoin.SKY_OK
    _, b = skycoin.SKY_cipher_RandByte(160)
    r = skycoin.cipher_Ripemd160()
    skycoin.SKY_cipher_HashRipemd160(b, r)
    assert r != skycoin.cipher_Ripemd160()

    _, b = skycoin.SKY_cipher_RandByte(256)
    r2 = skycoin.cipher_Ripemd160()
    skycoin.SKY_cipher_HashRipemd160(b, r2)
    assert r2 != skycoin.cipher_Ripemd160()
    assert r2 == freshSumRipemd160(b)


def test_TestRipemd160Set():
    h = skycoin.cipher_Ripemd160()
    _, b = skycoin.SKY_cipher_RandByte(21)
    assert skycoin.SKY_cipher_Ripemd160_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthRipemd160
    _, b = skycoin.SKY_cipher_RandByte(100)
    assert skycoin.SKY_cipher_Ripemd160_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthRipemd160
    _, b = skycoin.SKY_cipher_RandByte(19)
    assert skycoin.SKY_cipher_Ripemd160_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthRipemd160
    _, b = skycoin.SKY_cipher_RandByte(0)
    assert skycoin.SKY_cipher_Ripemd160_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthRipemd160
    _, b = skycoin.SKY_cipher_RandByte(20)
    assert skycoin.SKY_cipher_Ripemd160_Set(h, b) == skycoin.SKY_OK
    _, b1 = skycoin.SKY_cipher_RandByte(20)
    skycoin.SKY_cipher_Ripemd160_Set(h, b1)
    assert h.compareToString(str(b))


def test_TestSHA256Set():
    h = skycoin.cipher_SHA256()
    _, b = skycoin.SKY_cipher_RandByte(33)
    assert skycoin.SKY_cipher_SHA256_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthSHA256
    _, b = skycoin.SKY_cipher_RandByte(100)
    assert skycoin.SKY_cipher_SHA256_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthSHA256
    _, b = skycoin.SKY_cipher_RandByte(31)
    assert skycoin.SKY_cipher_SHA256_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthSHA256
    _, b = skycoin.SKY_cipher_RandByte(0)
    assert skycoin.SKY_cipher_SHA256_Set(
        h, b) == skycoin.SKY_ErrInvalidLengthSHA256
    _, b = skycoin.SKY_cipher_RandByte(32)
    assert skycoin.SKY_cipher_SHA256_Set(h, b) == skycoin.SKY_OK
    _, b = skycoin.SKY_cipher_RandByte(32)
    skycoin.SKY_cipher_SHA256_Set(h, b)
    assert h.toStr()[:] == b


def test_TestSHA256Hex():
    h = skycoin.cipher_SHA256()
    _, b = skycoin.SKY_cipher_RandByte(32)
    skycoin.SKY_cipher_SHA256_Set(h, b)
    _, s = skycoin.SKY_cipher_SHA256_Hex(h)
    h2 = skycoin.cipher_SHA256()
    err = skycoin.SKY_cipher_SHA256FromHex(s, h2)
    assert err == skycoin.SKY_OK
    assert h == h2
    _, s2 = skycoin.SKY_cipher_SHA256_Hex(h2)
    assert s2 == s


class struct_test:
    inputs = bytes
    outputs = bytes


def test_TestSHA256KnownValue():
    vals = []
    values = struct_test()
    values.inputs = b"skycoin"
    values.outputs = b"5a42c0643bdb465d90bf673b99c14f5fa02db71513249d904573d2b8b63d353d"
    vals.append(values)
    values.inputs = b"hello world"
    values.outputs = b"b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9"
    vals.append(values)
    values.inputs = b"hello world asd awd awd awdapodawpokawpod "
    values.outputs = b"99d71f95cafe05ea2dddebc35b6083bd5af0e44850c9dc5139b4476c99950be4"

    for v in vals:
        b = v.inputs
        h = skycoin.cipher_SHA256()
        err = skycoin.SKY_cipher_SumSHA256(b, h)
        assert err == skycoin.SKY_OK
        _, h = skycoin.SKY_cipher_SHA256_Hex(h)
        assert h == v.outputs


def test_TestSumSHA256():
    _, b = skycoin.SKY_cipher_RandByte(256)
    h1 = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(b, h1)
    assert h1 != skycoin.cipher_SHA256()

    _, c = skycoin.SKY_cipher_RandByte(256)
    h2 = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(c, h2)
    assert h2 != skycoin.cipher_SHA256()
    assert h2 == freshSumSHA256(c)


def test_TestSHA256FromHex():
    # Invalid hex hash
    h = skycoin.cipher_SHA256()
    err = skycoin.SKY_cipher_SHA256FromHex(b"cawcad", h)
    assert err == skycoin.SKY_ERROR

    # Truncated hex hash
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(b, h)
    h_bytes = h.toStr()
    h1 = skycoin.cipher_SHA256()
    err = skycoin.SKY_cipher_SHA256FromHex(h_bytes[:int(len(h_bytes) / 2)], h1)
    assert err == skycoin.SKY_ERROR

    # Valid hex hash
    h2 = skycoin.cipher_SHA256()
    err, b = skycoin.SKY_cipher_SHA256_Hex(h)
    err = skycoin.SKY_cipher_SHA256FromHex(b, h2)
    assert h == h2
    assert err == skycoin.SKY_OK


def test_TestDoubleSHA256():
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_DoubleSHA256(b, h)
    assert h != skycoin.cipher_SHA256()
    assert h != freshSumSHA256(b)


def test_TestAddSHA256():
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(b, h)
    _, c = skycoin.SKY_cipher_RandByte(64)
    i = skycoin.cipher_SHA256()
    skycoin.SKY_cipher_SumSHA256(c, i)
    add = skycoin.cipher_SHA256()
    err = skycoin.SKY_cipher_AddSHA256(h, i, add)
    assert err == skycoin.SKY_OK
    assert add != skycoin.cipher_SHA256()
    assert add != h
    assert add != i


def test_TestXorSHA256():
    _, b = skycoin.SKY_cipher_RandByte(128)
    _, c = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    i = skycoin.cipher_SHA256()
    err = skycoin.SKY_cipher_SumSHA256(b, h)
    assert err == skycoin.SKY_OK
    err = skycoin.SKY_cipher_SumSHA256(c, i)
    assert err == skycoin.SKY_OK
    temp = skycoin.cipher_SHA256()
    temp2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_cipher_SHA256_Xor(h, i, temp) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_SHA256_Xor(i, h, temp2) == skycoin.SKY_OK
    assert temp != h
    assert temp != i
    assert temp != skycoin.cipher_SHA256()
    assert temp == temp2


def test_TestSHA256Null():
    x = skycoin.cipher_SHA256()
    _, isNull = skycoin.SKY_cipher_SHA256_Null(x)
    assert isNull
    _, b = skycoin.SKY_cipher_RandByte(128)
    skycoin.SKY_cipher_SumSHA256(b, x)
    _, isNull = skycoin.SKY_cipher_SHA256_Null(x)
    assert not isNull


def test_TestMerkle():
    hashlist = []
    h = skycoin.cipher_SHA256()
    for _ in range(5):
        hashlist.append(h)

    for i in range(5):
        err, data = skycoin.SKY_cipher_RandByte(128)
        assert err == skycoin.SKY_OK
        err = skycoin.SKY_cipher_SumSHA256(data, hashlist[i])
        assert err == skycoin.SKY_OK

    # assert skycoin.SKY_cipher_Merkle(hashlist,h) == 45


# err, data = skycoin.SKY_cipher_RandByte(128)
# assert err == skycoin.SKY_OK
# h = skycoin.cipher_SHA256()
# err = skycoin.SKY_cipher_SumSHA256(data, h)
# assert err == skycoin.SKY_OK
# Single hash input returns hash
# a = [x for x in h]
# assert skycoin.SKY_cipher_Merkle(a,h) == 45
#  DoubleSHA256 double SHA256
# func DoubleSHA256(b []byte) SHA256 {
# 	h1 := SumSHA256(b)
# 	h2 := SumSHA256(h1[:])
# 	return h2
# }

# // AddSHA256 returns the SHA256 hash of to two concatenated hashes
# func AddSHA256(a SHA256, b SHA256) SHA256 {
# 	c := append(a[:], b[:]...)
# 	return SumSHA256(c)
# }

# // Returns the next highest power of 2 above n, if n is not already a
# // power of 2
# func nextPowerOfTwo(n uint64) uint64 {
# 	var k uint64 = 1
# 	for k < n {
# 		k *= 2
# 	}
# 	return k
# }

# // Merkle computes the merkle root of a hash array
# // Array of hashes is padded with 0 hashes until next power of 2
# func Merkle(h0 []SHA256) SHA256 {
# 	lh := uint64(len(h0))
# 	np := nextPowerOfTwo(lh)
# 	h1 := append(h0, make([]SHA256, np-lh)...)
# 	for len(h1) != 1 {
# 		h2 := make([]SHA256, len(h1)/2)
# 		for i := 0; i < len(h2); i++ {
# 			h2[i] = AddSHA256(h1[2*i], h1[2*i+1])
# 		}
# 		h1 = h2
# 	}
# 	return h1[0]
# }

# def merkle(h0):
#     lh = base64.standard_b64decode(len(h0))
#     np =

# def nextPowerOfTwo():



import skycoin
import tests.utils as utils


class tmpstruct:
    s = "",
    n = 0,
    e = 0

    def __init__(self, _s, _n, _e=0):
        self.e = _e
        self.s = _s
        self.n = _n


def test_TestFromString():
    cases = []
    cases.append(tmpstruct(b"0", 0))
    cases.append(tmpstruct(b"0.", 0))
    cases.append(tmpstruct(b"0.0", 0))
    cases.append(tmpstruct(b"0.000000", 0))
    cases.append(tmpstruct(b"0.0000000", 0))
    cases.append(tmpstruct(b"0.0000001", 0, skycoin.SKY_ErrTooManyDecimals))
    cases.append(tmpstruct(b"0.000001", 1))
    cases.append(tmpstruct(b"0.0000010", 1))
    cases.append(tmpstruct(b"1", int(1e6)))
    cases.append(tmpstruct(b"1.000001", int(1e6 + 1)))
    cases.append(tmpstruct(b"-1", 0, skycoin.SKY_ErrNegativeValue))
    cases.append(tmpstruct(b"10000", int(1e4 * 1e6)))
    cases.append(tmpstruct(b"123456789.123456", int(123456789123456)))
    cases.append(tmpstruct(b"123.000456", int(123000456)))
    cases.append(tmpstruct(b"100SKY", 0, skycoin.SKY_ERROR))
    cases.append(tmpstruct(b"", 0, skycoin.SKY_ERROR))
    cases.append(
        tmpstruct(
            b"999999999999999999999999999999999999999999",
            0,
            skycoin.SKY_ErrTooLarge))
    cases.append(tmpstruct(b"9223372036854.775807", 9223372036854775807))
    cases.append(
        tmpstruct(
            b"-9223372036854.775807",
            0,
            skycoin.SKY_ErrNegativeValue))
    cases.append(tmpstruct(b"9223372036854775808", 0, skycoin.SKY_ErrTooLarge))
    cases.append(
        tmpstruct(
            b"9223372036854775807.000001",
            0,
            skycoin.SKY_ErrTooLarge))
    cases.append(tmpstruct(b"9223372036854775807", 0, skycoin.SKY_ErrTooLarge))
    cases.append(
        tmpstruct(
            b"9223372036854775806.000001",
            0,
            skycoin.SKY_ErrTooLarge))
    cases.append(tmpstruct(b"1.1", int(1e6 + 1e5)))
    cases.append(tmpstruct(b"1.01", int(1e6 + 1e4)))
    cases.append(tmpstruct(b"1.001", int(1e6 + 1e3)))
    cases.append(tmpstruct(b"1.0001", int(1e6 + 1e2)))
    cases.append(tmpstruct(b"1.00001", int(1e6 + 1e1)))
    cases.append(tmpstruct(b"1.000001", int(1e6 + 1e0)))
    cases.append(tmpstruct(b"1.0000001", 0, skycoin.SKY_ErrTooManyDecimals))

    for i in range(len(cases)):
        tc = cases[i]
        err, n = skycoin.SKY_droplet_FromString(tc.s)

        if tc.e == skycoin.SKY_OK:
            assert tc.e == err
            assert tc.n == n
        else:
            assert tc.e == err
            assert 0 == n


def test_TestToString():
    cases = []
    cases.append(tmpstruct(b"0.000000", 0))
    cases.append(tmpstruct(b"0.000001", 1))
    cases.append(tmpstruct(b"1.000000", int(1e6)))
    cases.append(tmpstruct(b"0.100100", 100100))
    cases.append(tmpstruct(b"1.001000", 1001000))
    cases.append(tmpstruct(b"0.000999", 999))
    cases.append(tmpstruct(b"999.000000", 999000000))
    cases.append(tmpstruct(b"123.000456", 123000456))
    cases.append(tmpstruct(b"", 9223372036854775808, skycoin.SKY_ErrTooLarge))

    for tcc in cases:
        tc = tcc
        err, s = skycoin.SKY_droplet_ToString(tc.n)

        if tc.e == skycoin.SKY_OK:
            assert tc.s == s
        else:
            assert tc.e == err
            assert None == s



import skycoin
import tests.utils as utils


def test_TestTransactionVerify():
    # Mismatch header hash
    handle, tx = utils.makeTransaction()
    h = skycoin.cipher_SHA256()
    h.assignTo(tx.InnerHash)
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # No inputs
    handle, tx = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_ResetInputs(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # No outputs
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_ResetOutputs(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Invalid number of Sigs
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_ResetSignatures(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    assert skycoin.SKY_coin_Transaction_ResetSignatures(
        handle, 20) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Too many sigs & inputs
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_ResetSignatures(
        handle, utils.MaxUint16) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_ResetInputs(
        handle, utils.MaxUint16) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Duplicate inputs
    ux, s = utils.makeUxOutWithSecret()
    handle, _ = utils.makeTransactionFromUxOut(ux, s)
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_GetInputAt(
        handle, 0, h) == skycoin.SKY_OK
    r = skycoin.SKY_coin_Transaction_PushInput(handle, h)
    assert r == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_ResetSignatures(
        handle, 0) == skycoin.SKY_OK
    secKeys = []
    secKeys.append(s)
    secKeys.append(s)
    assert skycoin.SKY_coin_Transaction_SignInputs(
        handle, secKeys) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Duplicate outputs
    handle, _ = utils.makeTransaction()
    pOutput = skycoin.coin__TransactionOutput()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOutput) == skycoin.SKY_OK
    pOutput.Address = skycoin.cipher__Address()
    assert skycoin.SKY_coin_Transaction_ResetOutputs(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, pOutput.Coins, pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, pOutput.Coins, pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Output coins are 0
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOutput) == skycoin.SKY_OK
    pOutput.Coins = 0
    assert skycoin.SKY_coin_Transaction_ResetOutputs(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, pOutput.Coins, pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Output coin overflow
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOutput) == skycoin.SKY_OK
    pOutput.Coins = int(utils.MaxUint64 - int(3e6))
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, pOutput.Coins, pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_ERROR
    # Output coins are not multiples of 1e6 (valid, decimal restriction is not
    # enforced here)
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOutput) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_ResetOutputs(
        handle, 0) == skycoin.SKY_OK
    pOutput.Coins += 10
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, pOutput.Coins, pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_ResetSignatures(
        handle, 0) == skycoin.SKY_OK
    p = skycoin.cipher_PubKey()
    s = skycoin.cipher_SecKey()
    assert skycoin.SKY_cipher_GenerateKeyPair(p, s) == skycoin.SKY_OK
    secKeys = []
    secKeys.append(s)
    assert skycoin.SKY_coin_Transaction_SignInputs(
        handle, secKeys) == skycoin.SKY_OK
    assert int(pOutput.Coins % int(1e6)) != int(0)
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_OK
    # Valid
    handle, _ = utils.makeTransaction()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOutput) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_ResetOutputs(
        handle, 0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, int(10e6), pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, pOutput.Address, int(1e6), pOutput.Hours) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_Verify(handle) == skycoin.SKY_OK


def test_TestTransactionVerifyInput():
    # Valid
    ux, s = utils.makeUxOutWithSecret()
    handle, tx = utils.makeTransactionFromUxOut(ux, s)
    seckeys = []
    seckeys.append(ux)
    assert skycoin.SKY_coin_VerifyInputSignatures(
        handle, seckeys) == skycoin.SKY_OK


def test_TestTransactionPushInput():
    handle = utils.makeEmptyTransaction()
    ux = utils.makeUxOut()
    sha = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_Hash(ux, sha) == skycoin.SKY_OK
    r = skycoin.SKY_coin_Transaction_PushInput(handle, sha)
    assert r == 0
    _, count = skycoin.SKY_coin_Transaction_GetInputsCount(handle)
    assert count == 1
    sha1 = skycoin.cipher_SHA256()
    skycoin.SKY_coin_Transaction_GetInputAt(handle, 0, sha1)
    assert sha == sha1
    skycoin.SKY_coin_Transaction_ResetInputs(handle, 0)
    for _ in range(utils.MaxUint16):
        skycoin.SKY_coin_Transaction_PushInput(
            handle, skycoin.cipher_SHA256())
    ux = utils.makeUxOut()
    assert skycoin.SKY_coin_UxOut_Hash(ux, sha) == skycoin.SKY_OK


def test_TestTransactionPushOutput():
    handle = utils.makeEmptyTransaction()
    a = utils.makeAddress()
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, a, 100, 150) == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transaction_GetOutputsCount(handle)
    assert err == skycoin.SKY_OK
    assert count == 1
    pOut1 = skycoin.coin__TransactionOutput()
    pOut = skycoin.coin__TransactionOutput()
    pOut1.Address = a
    pOut1.Coins = 100
    pOut1.Hours = 150
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle, 0, pOut) == skycoin.SKY_OK
    assert pOut == pOut1
    for i in range(1, 20):
        a = utils.makeAddress()
        assert skycoin.SKY_coin_Transaction_PushOutput(
            handle, a, int(i * 100), int(i * 50)) == skycoin.SKY_OK
        err, count = skycoin.SKY_coin_Transaction_GetOutputsCount(handle)
        assert err == skycoin.SKY_OK
        assert count == int(i + 1)
        pOut1.Address = a
        pOut1.Coins = int(i * 100)
        pOut1.Hours = int(i * 150)
        assert skycoin.SKY_coin_Transaction_GetOutputAt(
            handle, i, pOut) == skycoin.SKY_OK
        assert pOut == pOut
        i += 1


def test_TestTransactionSignInputs():
    handle = utils.makeEmptyTransaction()
    # Panics if txns already signed
    sig = skycoin.cipher_Sig()
    assert skycoin.SKY_coin_Transaction_PushSignature(
        handle, sig) == skycoin.SKY_OK
    secKeys = []
    secKeys.append(skycoin.cipher_SecKey())
    # Panics if not enough keys
    handle = utils.makeEmptyTransaction()
    ux, s = utils.makeUxOutWithSecret()
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_Hash(ux, h) == skycoin.SKY_OK
    skycoin.SKY_coin_Transaction_PushInput(handle, h)
    ux2, s2 = utils.makeUxOutWithSecret()
    assert skycoin.SKY_coin_UxOut_Hash(ux2, h) == skycoin.SKY_OK
    skycoin.SKY_coin_Transaction_PushInput(handle, h)
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), 40, 80) == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transaction_GetSignaturesCount(handle)
    assert err == skycoin.SKY_OK
    assert count == 0
    # Valid signing
    assert skycoin.SKY_coin_Transaction_HashInner(handle, h) == skycoin.SKY_OK
    secKeys = []
    secKeys.append(s)
    secKeys.append(s2)
    assert skycoin.SKY_coin_Transaction_SignInputs(
        handle, secKeys) == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transaction_GetSignaturesCount(handle)
    assert err == skycoin.SKY_OK
    assert count == 2
    h2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle, h2) == skycoin.SKY_OK
    assert h == h2
    p = skycoin.cipher_PubKey()
    assert skycoin.SKY_cipher_PubKeyFromSecKey(s, p) == skycoin.SKY_OK
    a = skycoin.cipher__Address()
    a2 = skycoin.cipher__Address()
    assert skycoin.SKY_cipher_AddressFromPubKey(p, a) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_PubKeyFromSecKey(s2, p) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_AddressFromPubKey(p, a2) == skycoin.SKY_OK
    sha1 = skycoin.cipher_SHA256()
    sha2 = skycoin.cipher_SHA256()
    txin0 = skycoin.cipher_SHA256()
    txin1 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_GetInputAt(
        handle, 0, txin0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_GetInputAt(
        handle, 1, txin1) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_AddSHA256(h, txin0, sha1) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_AddSHA256(h, txin1, sha2) == skycoin.SKY_OK
    txsig0 = skycoin.cipher_Sig()
    txsig1 = skycoin.cipher_Sig()
    assert skycoin.SKY_coin_Transaction_GetSignatureAt(
        handle, 0, txsig0) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_GetSignatureAt(
        handle, 1, txsig1) == skycoin.SKY_OK


def test_TestTransactionHash():
    handle, _ = utils.makeTransaction()
    h = skycoin.cipher_SHA256()
    h2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_Hash(handle, h) == skycoin.SKY_OK
    assert h != h2
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle, h2) == skycoin.SKY_OK
    assert h != h2


def test_TestTransactionUpdateHeader():
    handle, tx = utils.makeTransaction()
    h = skycoin.cipher_SHA256()
    h1 = skycoin.cipher_SHA256()
    h2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_HashInner(handle, h) == skycoin.SKY_OK
    skycoin.cipher_SHA256().assignTo(tx.InnerHash)
    assert skycoin.SKY_coin_Transaction_UpdateHeader(handle) == skycoin.SKY_OK
    h1.assignFrom(tx.InnerHash)
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle, h2) == skycoin.SKY_OK
    assert h1 != skycoin.cipher_SHA256()
    assert h1 == h
    assert h1 == h2


def test_TestTransactionHashInner():
    handle, tx = utils.makeTransaction()
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_HashInner(handle, h) == skycoin.SKY_OK
    assert h != skycoin.cipher_SHA256()

    #  If tx.In is changed, hash should change
    handle2, tx2 = utils.copyTransaction(handle)
    ux = utils.makeUxOut()
    h = skycoin.cipher_SHA256()
    h1 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_Hash(ux, h) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_SetInputAt(
        handle2, 0, h) == skycoin.SKY_OK
    assert tx != tx2
    assert skycoin.SKY_coin_UxOut_Hash(ux, h1) == skycoin.SKY_OK
    assert h == h1
    assert skycoin.SKY_coin_Transaction_HashInner(handle, h) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle2, h1) == skycoin.SKY_OK
    assert h != h1

    # If tx.Out is changed, hash should change
    handle2, tx2 = utils.copyTransaction(handle)
    a = utils.makeAddress()
    a2 = skycoin.cipher__Address()
    pOut = skycoin.coin__TransactionOutput()
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle2, 0, pOut) == skycoin.SKY_OK
    pOut.Address = a
    assert skycoin.SKY_coin_Transaction_SetOutputAt(
        handle2, 0, pOut) == skycoin.SKY_OK
    assert tx != tx2
    assert skycoin.SKY_coin_Transaction_GetOutputAt(
        handle2, 0, pOut) == skycoin.SKY_OK
    assert pOut.Address == a
    sha1 = skycoin.cipher_SHA256()
    sha2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle, sha1) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle2, sha2) == skycoin.SKY_OK
    assert sha1 != sha2

    # If tx.Head is changed, hash should not change
    handle2, tx2 = utils.copyTransaction(handle)
    sig = skycoin.cipher_Sig()
    assert skycoin.SKY_coin_Transaction_PushSignature(
        handle, sig) == skycoin.SKY_OK
    sha1 = skycoin.cipher_SHA256()
    sha2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle, sha1) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_HashInner(
        handle2, sha2) == skycoin.SKY_OK
    assert sha1 == sha2


def test_TestTransactionSerialization():
    handle, tx = utils.makeTransaction()
    err, b = skycoin.SKY_coin_Transaction_Serialize(handle)
    assert err == skycoin.SKY_OK
    err, handle2 = skycoin.SKY_coin_TransactionDeserialize(b)
    assert err == skycoin.SKY_OK
    err, tx2 = skycoin.SKY_coin_GetTransactionObject(handle2)
    assert err == skycoin.SKY_OK
    assert tx == tx2


def test_TestTransactionOutputHours():
    handle = utils.makeEmptyTransaction()
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), int(1e6), 100) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), int(1e6), 200) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), int(1e6), 500) == skycoin.SKY_OK
    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), int(1e6), 0) == skycoin.SKY_OK
    err, hours = skycoin.SKY_coin_Transaction_OutputHours(handle)
    assert err == skycoin.SKY_OK
    assert hours == 800

    assert skycoin.SKY_coin_Transaction_PushOutput(
        handle, utils.makeAddress(), int(1e6), int(
            utils.MaxUint64 - 700)) == skycoin.SKY_OK
    err, _ = skycoin.SKY_coin_Transaction_OutputHours(handle)
    assert err == skycoin.SKY_ERROR


def test_TestTransactionsSize():
    handle = utils.makeTransactions(10)
    size = 0
    for i in range(10):
        err, tx = skycoin.SKY_coin_Transactions_GetAt(handle, i)
        assert err == skycoin.SKY_OK
        err, b = skycoin.SKY_coin_Transaction_Serialize(tx)
        size += len(b)
        i += 1

    assert size != 0
    err, sizetx = skycoin.SKY_coin_Transactions_Size(handle)
    assert err == skycoin.SKY_OK
    assert sizetx == size


def test_TestTransactionsHashes():
    handle = utils.makeTransactions(4)
    err, hashes = skycoin.SKY_coin_Transactions_Hashes(handle)
    assert err == skycoin.SKY_OK
    len_hashes = len(hashes)
    assert len_hashes == 4
    for i in range(len_hashes):
        err, tx = skycoin.SKY_coin_Transactions_GetAt(handle, i)
        assert err == skycoin.SKY_OK
        h = skycoin.cipher_SHA256()
        assert skycoin.SKY_coin_Transaction_Hash(tx, h) == skycoin.SKY_OK
        assert h == hashes[i]
        i += 1


def test_TestTransactionsTruncateBytesTo():
    handles = utils.makeTransactions(10)
    trunc = 0
    for i in range(5):
        err, handle = skycoin.SKY_coin_Transactions_GetAt(handles, i)
        assert err == skycoin.SKY_OK
        err, count = skycoin.SKY_coin_Transaction_Size(handle)
        assert err == skycoin.SKY_OK
        trunc += count
        i += 1
    # Trucating halfway
    err, tnxs2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert err == skycoin.SKY_OK
    err, len_tnxs2 = skycoin.SKY_coin_Transactions_Length(tnxs2)
    assert err == skycoin.SKY_OK
    assert len_tnxs2 == 5
    err, count = skycoin.SKY_coin_Transactions_Size(tnxs2)
    assert err == skycoin.SKY_OK
    assert count == trunc

    # Stepping into next boundary has same cutoff, must exceed
    trunc += 1
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Length(tnxs2)
    assert err == skycoin.SKY_OK
    assert count == 5
    err, count = skycoin.SKY_coin_Transactions_Size(tnxs2)
    assert err == skycoin.SKY_OK
    assert count == int(trunc - 1)

    # Moving to 1 before next level
    err, tnxs_5 = skycoin.SKY_coin_Transactions_GetAt(handles, 5)
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transaction_Size(tnxs_5)
    assert err == skycoin.SKY_OK
    trunc += int(count - 2)
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Length(txns2)
    assert err == skycoin.SKY_OK
    assert count == 5
    err, count = skycoin.SKY_coin_Transactions_Size(txns2)
    assert err == skycoin.SKY_OK
    err, count_tnxs5 = skycoin.SKY_coin_Transaction_Size(tnxs_5)
    assert err == skycoin.SKY_OK
    assert int(trunc - count_tnxs5 + 1) == count

    # Moving to next level
    trunc += 1
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Length(txns2)
    assert err == skycoin.SKY_OK
    assert count == 6
    err, count = skycoin.SKY_coin_Transactions_Size(txns2)
    assert err == skycoin.SKY_OK
    assert count == trunc

    # Truncating to full available amt
    err, trunc = skycoin.SKY_coin_Transactions_Size(handles)
    assert err == skycoin.SKY_OK
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert err == skycoin.SKY_OK
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Size(txns2)
    assert err == skycoin.SKY_OK
    assert count == trunc
    assert utils.equalTransactions(handles, txns2) == skycoin.SKY_OK

    # Truncating over amount
    trunc += 1
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, trunc)
    assert utils.equalTransactions(handles, txns2) == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Size(handles)
    assert err == skycoin.SKY_OK
    assert count == int(trunc - 1)

    # Truncating to 0
    trunc = 0
    err, txns2 = skycoin.SKY_coin_Transactions_TruncateBytesTo(handles, 0)
    assert err == skycoin.SKY_OK
    err, count = skycoin.SKY_coin_Transactions_Length(txns2)
    assert err == skycoin.SKY_OK
    assert count == 0
    err, count = skycoin.SKY_coin_Transactions_Size(txns2)
    assert err == skycoin.SKY_OK
    assert count == trunc


class ux():
    coins = 0
    hours = 0


class cases():
    name = ""
    inUxs = []
    outUxs = []
    err = 0
    headTime = 0


def test_TestVerifyTransactionCoinsSpending():
    case = []
    inu1 = ux()
    tests1 = cases
    tests1.name = "Input coins overflow"
    tests1.err = skycoin.SKY_ERROR
    inu1.coins = int(utils.MaxUint64 - int(1e6) + 1)
    inu1.hours = 10
    tests1.inUxs.append(inu1)
    inu1.coins = int(1e6)
    inu1.hours = 0
    tests1.inUxs.append(inu1)
    case.append(tests1)

    tests2 = cases
    inu2 = ux
    tests2.name = "Output coins overflow"
    tests2.err = skycoin.SKY_ERROR
    inu2.coins = int(10e6)
    inu2.hours = 10
    tests2.inUxs.append(inu2)
    inu2.coins = int(utils.MaxUint64 - 10e6 + 1)
    inu2.hours = 0
    tests2.outUxs.append(inu2)
    inu2.coins = int(20e6)
    inu2.hours = 1
    tests2.outUxs.append(inu2)
    case.append(tests2)

    tests3 = cases
    inu3 = ux
    tests3.name = "Insufficient coins"
    tests3.err = skycoin.SKY_ERROR
    inu3.coins = int(10e6)
    inu3.hours = int(10)
    tests3.inUxs.append(inu3)
    inu3.coins = int(15e6)
    inu3.hours = int(10)
    tests3.inUxs.append(inu3)
    inu3.coins = int(20e6)
    inu3.hours = int(1)
    tests3.outUxs.append(inu3)
    inu3.coins = int(10e6)
    inu3.hours = int(1)
    tests3.outUxs.append(inu3)
    case.append(tests3)

    tests4 = cases
    inu4 = ux
    tests4.name = "Destroyed coins"
    tests4.err = skycoin.SKY_ERROR
    inu4.coins = int(10e6)
    inu4.hours = int(10)
    tests4.inUxs.append(inu4)
    inu4.coins = int(15e6)
    inu4.hours = int(10)
    tests4.inUxs.append(inu4)
    inu4.coins = int(5e6)
    inu4.hours = int(1)
    tests4.outUxs.append(inu4)
    inu4.coins = int(10e6)
    inu4.hours = int(1)
    tests4.outUxs.append(inu4)
    case.append(tests4)

    tests5 = cases
    inu5 = ux
    tests5.name = "valid"

    inu5.coins = int(10e6)
    inu5.hours = int(10)
    tests5.inUxs.append(inu5)
    inu5.coins = int(15e6)
    inu5.hours = int(10)
    tests5.inUxs.append(inu5)

    inu5.coins = int(10e6)
    inu5.hours = int(11)
    tests5.outUxs.append(inu5)
    inu5.coins = int(10e6)
    inu5.hours = int(1)
    tests5.outUxs.append(inu5)
    inu5.coins = int(5e6)
    inu5.hours = int(0)
    tests5.outUxs.append(inu5)
    case.append(tests5)

    for tc in case:
        uxIn = []
        uxOut = []
        for ch in tc.inUxs:
            puxIn = skycoin.coin__UxOut()
            puxIn.Body.Coins = ch.coins
            puxIn.Body.Hours = ch.hours
            uxIn.append(puxIn)
        for ch in tc.outUxs:
            puxOut = skycoin.coin__UxOut()
            puxOut.Body.Coins = ch.coins
            puxOut.Body.Hours = ch.hours
            uxOut.append(puxOut)
        assert skycoin.SKY_coin_VerifyTransactionCoinsSpending(
            uxIn, uxOut) == tc.err


def test_TestVerifyTransactionHoursSpending():
    case = []
    # Case #1
    case1 = cases()
    case1.name = "Input hours overflow"
    ux1_1 = ux()
    ux1_1.coins = int(3e6)
    ux1_1.hours = int(utils.MaxUint64 - int(1e6) + 1)
    case1.inUxs.append(ux1_1)
    ux1_2 = ux()
    ux1_2.coins = int(1e6)
    ux1_2.hours = int(1e6)
    case1.inUxs.append(ux1_2)
    case1.err = skycoin.SKY_ERROR
    case.append(case1)

    # Case #2
    case2 = cases()
    case2.name = "Insufficient coin hours"
    case2.err = skycoin.SKY_ERROR
    ux2_1 = ux()
    ux2_1.coins = int(10e6)
    ux2_1.hours = 10
    case2.inUxs.append(ux2_1)
    ux2_2 = ux()
    ux2_2.coins = int(15e6)
    ux2_2.hours = 10
    case2.inUxs.append(ux2_2)
    ox2_1 = ux()
    ox2_1.coins = int(15e6)
    ox2_1.hours = 10
    case2.outUxs.append(ox2_1)
    ox2_2 = ux()
    ox2_2.coins = int(10e6)
    ox2_2.hours = 11
    case2.outUxs.append(ox2_2)
    case.append(case2)

    # Case #3
    case3 = cases()
    case3.name = "coin hours time calculation overflow"
    case3.err = skycoin.SKY_ERROR
    case3.headTime = utils.MaxUint64
    ux3_1 = ux()
    ux3_1.coins = int(10e6)
    ux3_1.hours = 10
    case3.inUxs.append(ux3_1)
    ux3_2 = ux()
    ux3_2.coins = int(15e6)
    ux3_2.hours = 10
    case3.inUxs.append(ux3_2)
    ox3_1 = ux()
    ox3_1.coins = int(10e6)
    ox3_1.hours = 11
    case3.outUxs.append(ox3_1)
    ox3_2 = ux()
    ox3_2.coins = int(10e6)
    ox3_2.hours = 1
    case3.outUxs.append(ox3_2)
    ox3_3 = ux()
    ox3_3.coins = int(5e6)
    ox3_3.hours = 0
    case3.outUxs.append(ox3_3)
    case.append(case3)

    # Case #4
    case4 = cases()
    case4.name = "Invalid (coin hours overflow when adding earned hours, which is treated as 0, and now enough coin hours)"
    case4.err = skycoin.SKY_ERROR
    case4.headTime = int(1e6)
    ux4_1 = ux()
    ux4_1.coins = int(10e6)
    ux4_1.hours = utils.MaxUint64
    case4.inUxs.append(ux4_1)
    ox4_1 = ux()
    ox4_1.coins = int(10e6)
    ox4_1.hours = 1
    case4.outUxs.append(ox4_1)

    # Case #5
    case5 = cases()
    case5.name = "Valid (coin hours overflow when adding earned hours, which is treated as 0, but not sending any hours)"
    case5.headTime = int(1e6)
    ux5_1 = ux()
    ux5_1.coins = int(10e6)
    ux5_1.hours = utils.MaxUint64
    case5.inUxs.append(ux5_1)
    ox5_1 = ux()
    ox5_1.coins = int(10e6)
    ox5_1.hours = 0
    case5.outUxs.append(ox5_1)
    case.append(case5)

    # Case #6
    case6 = cases()
    case6.err = skycoin.SKY_OK
    case6.name = "Valid (base inputs have insufficient coin hours, but have sufficient after adjusting coinhours by headTime)"
    case6.headTime = 1492707255
    ux6_1 = ux()
    ux6_1.coins = int(10e6)
    ux6_1.hours = 10
    case6.inUxs.append(ux6_1)
    ux6_2 = ux()
    ux6_2.coins = int(15e6)
    ux6_2.hours = 10
    case6.inUxs.append(ux6_2)
    ox6_1 = ux()
    ox6_1.coins = int(15e6)
    ox6_1.hours = 10
    case6.outUxs.append(ox6_1)
    ox6_2 = ux()
    ox6_2.coins = int(10e6)
    ox6_2.hours = 11
    case6.outUxs.append(ox6_2)
    case.append(case6)

    # Case #7
    case7 = cases()
    case7.name = "valid"
    ux7_1 = ux()
    ux7_1.coins = int(10e6)
    ux7_1.hours = 10
    case7.inUxs.append(ux7_1)
    ux7_2 = ux()
    ux7_2.coins = int(15e6)
    ux7_2.hours = 10
    case7.inUxs.append(ux7_2)
    ox7_1 = ux()
    ox7_1.coins = int(10e6)
    ox7_1.hours = 11
    case7.outUxs.append(ox7_1)
    ox7_2 = ux()
    ox7_2.coins = int(10e6)
    ox7_2.hours = 1
    case7.outUxs.append(ox7_2)
    ox7_3 = ux()
    ox7_3.coins = int(5e6)
    ox7_3.hours = 0
    case7.outUxs.append(ox7_3)
    case.append(case7)

    for tc in case:
        uxIn = []
        uxOut = []
        for ch in tc.inUxs:
            puxIn = skycoin.coin__UxOut()
            puxIn.Body.Coins = ch.coins
            puxIn.Body.Hours = ch.hours
            uxIn.append(puxIn)
        for ch in tc.outUxs:
            puxOut = skycoin.coin__UxOut()
            puxOut.Body.Coins = ch.coins
            puxOut.Body.Hours = ch.hours
            uxOut.append(puxOut)
        print(tc.name)
        assert skycoin.SKY_coin_VerifyTransactionHoursSpending(
            tc.headTime, uxIn, uxOut) == tc.err


def calc(transactions):
    return skycoin.SKY_OK, 1


def overflowCalc(transaction):
    return skycoin.SKY_ERROR, utils.MaxUint64


def test_TestTransactionsFees():
    txns = utils.makeTransactions(0)
    # Nil txns
    err, fee = skycoin.SKY_coin_Transactions_Fees(
        txns, calc)
    assert err == skycoin.SKY_OK
    assert fee == 0

    # 2 transactions, calc() always returns 1
    txn = utils.makeEmptyTransaction()
    assert skycoin.SKY_coin_Transactions_Add(txns, txn) == skycoin.SKY_OK
    txn1 = utils.makeEmptyTransaction()
    assert skycoin.SKY_coin_Transactions_Add(txns, txn1) == skycoin.SKY_OK
    err, fee = skycoin.SKY_coin_Transactions_Fees(
        txns, calc)
    assert err == skycoin.SKY_OK
    assert fee == 2

    # calc error
    err, _ = skycoin.SKY_coin_Transactions_Fees(
        txns, utils.badFeeCalculator)
    assert err == skycoin.SKY_ERROR

    # # summing of calculated fees overflows
    err, _ = skycoin.SKY_coin_Transactions_Fees(txns, overflowCalc)
    assert err == skycoin.SKY_ERROR



import skycoin
import tests.utils as utils
from random import randint

_testSeckey = [
    b"08efb79385c9a8b0d1c6f5f6511be0c6f6c2902963d874a3a4bacc18802528d3",
    b"78298d9ecdc0640c9ae6883201a53f4518055442642024d23c45858f45d0c3e6",
    b"04e04fe65bfa6ded50a12769a3bd83d7351b2dbff08c9bac14662b23a3294b9e",
    b"2f5141f1b75747996c5de77c911dae062d16ae48799052c04ead20ccd5afa113",

]


def test_Test_Abnormal_Keys2():
    for tt in _testSeckey:
        err, s = skycoin.SKY_base58_String2Hex(tt)
        assert err == skycoin.SKY_OK
        err, p = skycoin.SKY_secp256k1_PubkeyFromSeckey(s)
        assert err == skycoin.SKY_OK
        err = skycoin.SKY_secp256k1_VerifyPubkey(p)


def test_Test_Abnormal_Keys3():
    for tt in _testSeckey:
        err, s = skycoin.SKY_base58_String2Hex(tt)
        assert err == skycoin.SKY_OK
        err, p = skycoin.SKY_secp256k1_PubkeyFromSeckey(s)
        assert err == skycoin.SKY_OK
        n = randint(0, len(_testSeckey) - 1)
        err, s2 = skycoin.SKY_base58_String2Hex(_testSeckey[n])
        assert err == skycoin.SKY_OK
        err, p2 = skycoin.SKY_secp256k1_PubkeyFromSeckey(s2)
        assert err == skycoin.SKY_OK
        err, puba = skycoin.SKY_secp256k1_ECDH(p, s2)
        assert err == skycoin.SKY_OK
        err, pubb = skycoin.SKY_secp256k1_ECDH(p2, s)
        assert err == skycoin.SKY_OK
        assert puba == pubb



import skycoin
import base64
import binascii

# InputTestData contains hashes to be signed


class InputTestData():
    Hashes = []


# SeedTestData contains data generated by Seed
class SeedTestData():
    Seed = []
    Keys = []


# KeysTestData contains address, public key,  secret key and list of signatures
class KeysTestData():
    Address = ""
    Secret = ""
    Public = ""
    Signatures = []


# InputTestDataFromJSON converts InputTestDataJSON to InputTestData
def InputTestDataFromJSON(InputTestDataJSON):
    hashes = []
    for h in InputTestDataJSON.Hashes:
        h2 = skycoin.cipher_SHA256()
        err = skycoin.SKY_cipher_SHA256FromHex(h.encode(), h2)  # ok python2
        hashes.append(h2)
        if err != skycoin.SKY_OK:
            return skycoin.SKY_ERROR, None
    return skycoin.SKY_OK, hashes


# KeysTestDataFromJSON converts KeysTestDataJSON to KeysTestData
def KeysTestDataFromJSON(KeysTestDataJSON):
    address = skycoin.cipher__Address()
    err = skycoin.SKY_cipher_DecodeBase58Address(
        KeysTestDataJSON["address"].encode(), address)
    if err != skycoin.SKY_OK:
        return skycoin.SKY_ERROR, None
    err, hex_str = skycoin.SKY_base58_String2Hex(
        KeysTestDataJSON["secret"].encode())
    assert err == skycoin.SKY_OK
    secret_key = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_NewSecKey(hex_str, secret_key)
    assert err == skycoin.SKY_OK

    err, secret_key_hex = skycoin.SKY_cipher_SecKey_Hex(secret_key)
    if err != skycoin.SKY_OK:
        return skycoin.SKY_ERROR, None

    err, hex_str = skycoin.SKY_base58_String2Hex(
        KeysTestDataJSON["public"].encode())
    assert err == skycoin.SKY_OK
    public_key = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_NewPubKey(hex_str, public_key)
    assert err == skycoin.SKY_OK

    err, public_key_hex = skycoin.SKY_cipher_PubKey_Hex(public_key)
    if err != skycoin.SKY_OK:
        return skycoin.SKY_ERROR, None

    r = KeysTestData()
    r.Address = address
    r.Public = public_key_hex
    r.Secret = secret_key_hex
    r.Signatures = 0
    if KeysTestDataJSON.get("signatures") is None:
        return skycoin.SKY_OK, r
    sigs = []
    if len(KeysTestDataJSON["signatures"]) >= 0:
        for s in KeysTestDataJSON["signatures"]:
            sig_fromHex = skycoin.cipher_Sig()
            err = skycoin.SKY_cipher_SigFromHex(s.encode(), sig_fromHex)
            assert err == skycoin.SKY_OK
            sigs.append(sig_fromHex)
            assert err == skycoin.SKY_OK
    r.Signatures = sigs
    return skycoin.SKY_OK, r


# SeedTestDataFromJSON converts SeedTestDataJSON to SeedTestData
def SeedTestDataFromJSON(SeedTestDataJSON):
    seed = base64.standard_b64decode(SeedTestDataJSON.Seed)
    keys = []
    for kj in SeedTestDataJSON.Keys:
        err, k = KeysTestDataFromJSON(kj)
        if err == skycoin.SKY_ERROR:
            return skycoin.SKY_ERROR, None
        keys.append(k)
    r = SeedTestData()
    r.Seed = seed
    r.Keys = keys
    return skycoin.SKY_OK, r

# ValidateSeedData validates the provided SeedTestData against the current cipher library.
# inputData is required if SeedTestData contains signatures


def ValidateSeedData(SeedTestData=None, InputTestData=None):
    err, keys = skycoin.SKY_cipher_GenerateDeterministicKeyPairs(
        SeedTestData.Seed, len(SeedTestData.Keys))
    assert err == skycoin.SKY_OK
    if len(SeedTestData.Keys) != len(keys):
        return skycoin.SKY_ERROR
    for i, s in enumerate(keys):
        secret_Key_null = skycoin.cipher_SecKey()
        if s == secret_Key_null:
            return skycoin.SKY_ErrInvalidSecKey
        if (SeedTestData.Keys[i].Secret).decode() != binascii.hexlify(
                bytearray(s.toStr())).decode('ascii'):
            assert err == skycoin.SKY_ERROR
        p = skycoin.cipher_PubKey()
        p_null = skycoin.cipher_PubKey()
        err = skycoin.SKY_cipher_PubKeyFromSecKey(s, p)
        if p == p_null:
            return skycoin.SKY_ErrInvalidPubKey
        if (SeedTestData.Keys[i].Public).decode() != binascii.hexlify(
                bytearray(p.toStr())).decode('ascii'):
            return skycoin.SKY_ErrInvalidPubKey
        addr1 = skycoin.cipher__Address()
        addr_null = skycoin.cipher__Address()
        err = skycoin.SKY_cipher_AddressFromPubKey(p, addr1)
        assert err == skycoin.SKY_OK
        if addr1 == addr_null:
            return skycoin.SKY_ErrAddressInvalidPubKey
        if not(SeedTestData.Keys[i].Address == addr1):
            return skycoin.SKY_ErrAddressInvalidChecksum
        addr2 = skycoin.cipher__Address()
        err = skycoin.SKY_cipher_AddressFromSecKey(s, addr2)
        assert err == skycoin.SKY_OK
        if not(addr1 == addr2):
            return skycoin.SKY_ErrAddressInvalidChecksum

        if InputTestData is None and SeedTestData.Keys[i].Signatures != 0:
            return skycoin.SKY_ERROR

        if InputTestData is not None:
            if len(SeedTestData.Keys[i].Signatures) != len(InputTestData):
                return skycoin.SKY_ERROR

            for j in range(len(InputTestData)):
                sig = SeedTestData.Keys[i].Signatures[j]
                sig_null = skycoin.cipher_Sig()
                if sig == sig_null:
                    return skycoin.SKY_ERROR

                #err = skycoin.SKY_cipher_VerifySignedHash(sig, InputTestData[j])
                # if err != skycoin.SKY_OK:
                #	return skycoin.SKY_ERROR

                p2 = skycoin.cipher_PubKey()
                err = skycoin.SKY_cipher_PubKeyFromSig(
                    sig, InputTestData[j], p2)
                if err != skycoin.SKY_OK:
                    return skycoin.SKY_ERROR
                if not(p == p2):
                    return 1
                    return skycoin.SKY_ERROR

                sig2 = skycoin.cipher_Sig()
                skycoin.SKY_cipher_SignHash(InputTestData[j], s, sig2)
                if sig2 == sig_null:
                    return skycoin.SKY_ERROR
    return skycoin.SKY_OK



import skycoin
import tests.utils as utils


class verifyTxFeeTestCase:
    inputHours = 0
    outputHours = 0
    err = 0

    def __init__(self, _inputHours, _outputHours, _err=0):
        self.inputHours = _inputHours
        self.outputHours = _outputHours
        self.err = _err


def test_TestVerifyTransactionFee():
    burnFactor2verifyTxFeeTestCase = []

    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(0, 0, skycoin.SKY_ErrTxnNoFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(1, 0, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(1, 1, skycoin.SKY_ErrTxnNoFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(2, 0, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(2, 1, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(2, 2, skycoin.SKY_ErrTxnNoFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(3, 0, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(3, 1, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(3, 2, skycoin.SKY_ErrTxnInsufficientFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(3, 3, skycoin.SKY_ErrTxnNoFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(4, 0, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(4, 1, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(4, 2, skycoin.SKY_OK))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(4, 3, skycoin.SKY_ErrTxnInsufficientFee))
    burnFactor2verifyTxFeeTestCase.append(
        verifyTxFeeTestCase(4, 4, skycoin.SKY_ErrTxnNoFee))
    addr = utils.makeAddress()
    emptyTxn = utils.makeEmptyTransaction()
    err, hours = skycoin.SKY_coin_Transaction_OutputHours(emptyTxn)
    assert err == skycoin.SKY_OK
    assert 0 == hours

    #  A txn with no outputs hours and no coinhours burn fee is valid
    err = skycoin.SKY_fee_VerifyTransactionFee(emptyTxn, 0, 2)
    assert err == skycoin.SKY_ErrTxnNoFee

    # A txn with no outputs hours but with a coinhours burn fee is valid
    err = skycoin.SKY_fee_VerifyTransactionFee(emptyTxn, 100, 2)
    assert err == skycoin.SKY_OK

    txn = utils.makeEmptyTransaction()
    addr = utils.makeAddress()
    err = skycoin.SKY_coin_Transaction_PushOutput(txn, addr, 0, int(1e6))
    assert err == skycoin.SKY_OK
    err = skycoin.SKY_coin_Transaction_PushOutput(txn, addr, 0, int(3e6))
    assert err == skycoin.SKY_OK

    err, hours = skycoin.SKY_coin_Transaction_OutputHours(txn)
    assert err == skycoin.SKY_OK
    assert hours == int(4e6)

    # A txn with insufficient net coinhours burn fee is invalid
    err = skycoin.SKY_fee_VerifyTransactionFee(txn, 0, 2)
    assert err == skycoin.SKY_ErrTxnNoFee

    err = skycoin.SKY_fee_VerifyTransactionFee(txn, 1, 2)
    assert err == skycoin.SKY_ErrTxnInsufficientFee

    # A txn with sufficient net coinhours burn fee is valid
    err, hours = skycoin.SKY_coin_Transaction_OutputHours(txn)
    assert err == skycoin.SKY_OK
    err = skycoin.SKY_fee_VerifyTransactionFee(txn, hours, 2)
    assert err == skycoin.SKY_OK
    err, hours = skycoin.SKY_coin_Transaction_OutputHours(txn)
    assert err == skycoin.SKY_OK
    err = skycoin.SKY_fee_VerifyTransactionFee(txn, hours * 10, 2)
    assert err == skycoin.SKY_OK

    # fee + hours overflows
    err = skycoin.SKY_fee_VerifyTransactionFee(
        txn, utils.MaxUint64 - int(3e6), 2)
    assert err == skycoin.SKY_ERROR

    # txn has overflowing output hours
    err = skycoin.SKY_coin_Transaction_PushOutput(
        txn, addr, 0, int(utils.MaxUint64 - 1e6 - 3e6 + 1))
    assert err == skycoin.SKY_OK
    err = skycoin.SKY_fee_VerifyTransactionFee(txn, 10, 2)
    assert err == skycoin.SKY_ERROR

    cases = burnFactor2verifyTxFeeTestCase

    for tc in cases:
        txn = utils.makeEmptyTransaction()
        err = skycoin.SKY_coin_Transaction_PushOutput(
            txn, addr, 0, tc.outputHours)
        assert tc.inputHours >= tc.outputHours
        err = skycoin.SKY_fee_VerifyTransactionFee(
            txn, int(tc.inputHours - tc.outputHours), 2)
        assert tc.err == err


class requiredFeeTestCase:
    hours = 0
    fee = 0

    def __init__(self, _hours, _fee):
        self.hours = _hours
        self.fee = _fee


def test_TestRequiredFee():

    burnFactor2RequiredFeeTestCases = []

    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(0, 0))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(1, 1))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(2, 1))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(3, 2))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(4, 2))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(5, 3))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(6, 3))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(7, 4))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(998, 499))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(999, 500))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(1000, 500))
    burnFactor2RequiredFeeTestCases.append(requiredFeeTestCase(1001, 501))
    cases = burnFactor2RequiredFeeTestCases

    for tc in cases:
        err, fee = skycoin.SKY_fee_RequiredFee(tc.hours, 2)
        assert err == skycoin.SKY_OK
        assert tc.fee == fee
        err, remainingHours = skycoin.SKY_fee_RemainingHours(tc.hours, 2)
        assert err == skycoin.SKY_OK
        assert ((tc.hours - fee) == remainingHours)


class uxInput:
    time = 0
    coins = 0
    hours = 0

    def __init__(self, _time, _coins, _hours):
        self.hours = _hours
        self.time = _time
        self.coins = _coins


class tmpstruct:
    name = ""
    out = []
    ins = []
    headTime = 0
    fee = 0
    err = 0

    def __init__(self):
        self.name = ""
        self.out = []
        self.ins = []
        self.headTime = 0
        self.fee = 0
        self.err = 0


def test_TestTransactionFee():
    headTime = int(1000)
    nextTime = int(headTime + 3600)  # 1 hour later

    cases = []
    #  Test case with one output, one input
    case1 = tmpstruct()
    case1.fee = 5
    case1.headTime = 1000
    case1.out = [5]
    case1.ins.append(uxInput(headTime, 10e6, 10))
    cases.append(case1)

    # Test case with multiple outputs, multiple inputs
    case2 = tmpstruct()
    case2.fee = 0
    case2.headTime = 1000
    case2.out = [5, 7, 3]
    case2.ins.append(uxInput(headTime, int(10e6), 10))
    case2.ins.append(uxInput(headTime, int(10e6), 5))
    cases.append(case2)
    # # Test case with multiple outputs, multiple inputs, and some inputs have more CoinHours once adjusted for HeadTime
    case3 = tmpstruct()
    case3.fee = 8
    case3.headTime = 1000
    case3.out.append(5)
    case3.out.append(10)
    case3.ins.append(uxInput(nextTime, 10e6, 10))
    case3.ins.append(uxInput(nextTime, 8e6, 5))
    # Test case with insufficient coin hours
    case4 = tmpstruct()
    case4.err = skycoin.SKY_ErrTxnInsufficientCoinHours
    case4.out.append(5)
    case4.out.append(10)
    case4.out.append(1)
    case4.ins.append(uxInput(headTime, 10e6, 10))
    case4.ins.append(uxInput(headTime, 8e6, 5))
    # Test case with overflowing input hours
    case5 = tmpstruct()
    case5.err = skycoin.SKY_ERROR
    case5.out.append(0)
    case5.ins.append(uxInput(headTime, 10e6, 10))
    case5.ins.append(uxInput(headTime, 10e6, utils.MaxUint64 - 9))
    case5.headTime = 1000
    # Test case with overflowing output hours
    case6 = tmpstruct()
    case6.err = skycoin.SKY_ERROR
    case6.out.append(0)
    case6.out.append(10)
    case6.out.append(utils.MaxUint64 - 9)
    case6.ins.append(uxInput(headTime, 10e6, 10))
    case6.ins.append(uxInput(headTime, 10e6, 100))
    case6.headTime = 1000
    addr = utils.makeAddress()
    for j in range(len(cases)):
        tc = cases[j]
        tx = utils.makeEmptyTransaction()
        for h in tc.out:
            err = skycoin.SKY_coin_Transaction_PushOutput(tx, addr, 0, h)
            assert err == skycoin.SKY_OK

        inUxs = utils.makeUxArray(len(tc.ins))
        for i in range(len(tc.ins)):
            b = tc.ins[i]
            inUxs[i].Head.Time = b.time
            inUxs[i].Body.Coins = int(b.coins)
            inUxs[i].Body.Hours = int(b.hours)
        err, fee = skycoin.SKY_fee_TransactionFee(tx, int(tc.headTime), inUxs)
        assert err == tc.err
        assert tc.fee == fee



import skycoin
import tests.utils as utils


def test_TestNewPubKey():
    public_key = skycoin.cipher_PubKey()
    _, data = skycoin.SKY_cipher_RandByte(31)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidLengthPubKey
    _, data = skycoin.SKY_cipher_RandByte(32)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidLengthPubKey
    _, data = skycoin.SKY_cipher_RandByte(34)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidLengthPubKey
    _, data = skycoin.SKY_cipher_RandByte(0)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidLengthPubKey
    _, data = skycoin.SKY_cipher_RandByte(100)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidLengthPubKey
    _, data = skycoin.SKY_cipher_RandByte(33)
    assert skycoin.SKY_cipher_NewPubKey(
        data, public_key) == skycoin.SKY_ErrInvalidPubKey

    pubkey = skycoin.cipher_PubKey()
    seckey = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_GenerateKeyPair(pubkey, seckey)
    assert err == skycoin.SKY_OK
    ptemp = pubkey.toStr()
    pubkey2 = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_NewPubKey(ptemp, pubkey2)
    assert err == skycoin.SKY_OK
    assert pubkey == pubkey2


def test_TestPubKeyVerify():
    # Random bytes should not be valid, most of the time
    failed = False
    for _ in range(10):
        public_key = skycoin.cipher_PubKey()
        _, data = skycoin.SKY_cipher_RandByte(33)
        skycoin.SKY_cipher_NewPubKey(data, public_key)
        if skycoin.SKY_cipher_PubKey_Verify(public_key) is not None:
            failed = True
            break
    assert failed is True


def test_TestPubKeyVerifyNil():
    # Empty public key should not be valid
    public_key = skycoin.cipher_PubKey()
    assert skycoin.SKY_cipher_PubKey_Verify(
        public_key) == skycoin.SKY_ErrInvalidPubKey


def test_TestPubKeyVerifyDefault1():
    #  Generated pub key should be valid
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK


def test_TestPubKeyVerifyDefault2():
    for _ in range(1024):
        public_key = skycoin.cipher_PubKey()
        secret_key = skycoin.cipher_SecKey()
        skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
        assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK


def test_TestPubKeyToAddress():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    addres = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, addres)
    # func (self Address) Verify(key PubKey) error
    assert skycoin.SKY_cipher_Address_Verify(
        addres, public_key) == skycoin.SKY_OK
    # func DecodeBase58Address(addr string) (Address, error)
    _, addres_str = skycoin.SKY_cipher_Address_String(addres)
    assert skycoin.SKY_cipher_DecodeBase58Address(
        addres_str, addres) == skycoin.SKY_OK


def test_TestPubKeyToAddress2():
    for _ in range(1024):
        public_key = skycoin.cipher_PubKey()
        secret_key = skycoin.cipher_SecKey()
        skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
        addres = skycoin.cipher__Address()
        skycoin.SKY_cipher_AddressFromPubKey(public_key, addres)
        # func (self Address) Verify(key PubKey) error
        assert skycoin.SKY_cipher_Address_Verify(
            addres, public_key) == skycoin.SKY_OK
        # func DecodeBase58Address(addr string) (Address, error)
        _, addres_str = skycoin.SKY_cipher_Address_String(addres)
        assert skycoin.SKY_cipher_DecodeBase58Address(
            addres_str, addres) == skycoin.SKY_OK


def test_TestMustNewSecKey():
    secret_key = skycoin.cipher_SecKey()
    _, data = skycoin.SKY_cipher_RandByte(31)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(33)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(34)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(0)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(100)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(32)
    assert skycoin.SKY_cipher_NewSecKey(data, secret_key) == skycoin.SKY_OK
    assert secret_key.toStr() == data


def test_TestSecKeyVerify():
    # Empty secret key should not be valid
    secret_key = skycoin.cipher_SecKey()
    public_key = skycoin.cipher_PubKey()
    assert skycoin.SKY_cipher_SecKey_Verify(secret_key) is not None
    # Generated sec key should be valid
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK


def test_TestECDHonce():
    secret_key_1 = skycoin.cipher_SecKey()
    public_key_1 = skycoin.cipher_PubKey()
    secret_key_2 = skycoin.cipher_SecKey()
    public_key_2 = skycoin.cipher_PubKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key_1, secret_key_1)
    skycoin.SKY_cipher_GenerateKeyPair(public_key_2, secret_key_2)
    _, data_1 = skycoin.SKY_cipher_ECDH(public_key_2, secret_key_1)
    _, data_2 = skycoin.SKY_cipher_ECDH(public_key_1, secret_key_2)
    assert data_1 == data_2


def test_TestECDHloop():
    for _ in range(128):
        secret_key_1 = skycoin.cipher_SecKey()
        public_key_1 = skycoin.cipher_PubKey()
        secret_key_2 = skycoin.cipher_SecKey()
        public_key_2 = skycoin.cipher_PubKey()
        skycoin.SKY_cipher_GenerateKeyPair(public_key_1, secret_key_1)
        skycoin.SKY_cipher_GenerateKeyPair(public_key_2, secret_key_2)
        _, data_1 = skycoin.SKY_cipher_ECDH(public_key_2, secret_key_1)
        _, data_2 = skycoin.SKY_cipher_ECDH(public_key_1, secret_key_2)
        assert data_1 == data_2


def test_TestNewSig():
    sig = skycoin.cipher_Sig()
    _, data = skycoin.SKY_cipher_RandByte(64)
    assert skycoin.SKY_cipher_NewSig(
        data, sig) == skycoin.SKY_ErrInvalidLengthSig
    _, data = skycoin.SKY_cipher_RandByte(66)
    assert skycoin.SKY_cipher_NewSig(
        data, sig) == skycoin.SKY_ErrInvalidLengthSig
    _, data = skycoin.SKY_cipher_RandByte(67)
    assert skycoin.SKY_cipher_NewSig(
        data, sig) == skycoin.SKY_ErrInvalidLengthSig
    _, data = skycoin.SKY_cipher_RandByte(0)
    assert skycoin.SKY_cipher_NewSig(
        data, sig) == skycoin.SKY_ErrInvalidLengthSig
    _, data = skycoin.SKY_cipher_RandByte(100)
    assert skycoin.SKY_cipher_NewSig(
        data, sig) == skycoin.SKY_ErrInvalidLengthSig
    _, data = skycoin.SKY_cipher_RandByte(65)
    assert skycoin.SKY_cipher_NewSig(data, sig) == skycoin.SKY_OK
    assert sig.toStr() == data


def test_TestSignHash():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    addres = skycoin.cipher__Address()
    sha_sum = skycoin.cipher_SHA256()
    sig_1 = skycoin.cipher_Sig()
    sig_2 = skycoin.cipher_Sig()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    skycoin.SKY_cipher_AddressFromPubKey(public_key, addres)
    _, data = skycoin.SKY_cipher_RandByte(256)
    skycoin.SKY_cipher_SumSHA256(data, sha_sum)
    skycoin.SKY_cipher_SignHash(sha_sum, secret_key, sig_1)
    assert sig_1 != sig_2


def test_TestPubKeyFromSecKey():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    public_key_2 = skycoin.cipher_PubKey()
    skycoin.SKY_cipher_PubKeyFromSecKey(secret_key, public_key_2)
    assert public_key == public_key_2
    secret_key_2 = skycoin.cipher_SecKey()
    assert skycoin.SKY_cipher_PubKeyFromSecKey(
        secret_key_2, public_key) == skycoin.SKY_ErrPubKeyFromNullSecKey
    _, data = skycoin.SKY_cipher_RandByte(99)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey
    _, data = skycoin.SKY_cipher_RandByte(31)
    assert skycoin.SKY_cipher_NewSecKey(
        data, secret_key) == skycoin.SKY_ErrInvalidLengthSecKey


def test_TestPubKeyFromSig():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    sha_sum = skycoin.cipher_SHA256()
    _, data = skycoin.SKY_cipher_RandByte(256)
    skycoin.SKY_cipher_SumSHA256(data, sha_sum)
    sig_1 = skycoin.cipher_Sig()
    skycoin.SKY_cipher_SignHash(sha_sum, secret_key, sig_1)
    public_key_2 = skycoin.cipher_PubKey()
    assert skycoin.SKY_cipher_PubKeyFromSig(
        sig_1, sha_sum, public_key_2) == skycoin.SKY_OK
    assert public_key == public_key_2
    sig_2 = skycoin.cipher_Sig()
    assert skycoin.SKY_cipher_PubKeyFromSig(
        sig_2, sha_sum, public_key_2) == skycoin.SKY_ErrInvalidSigPubKeyRecovery


def test_TestGenerateKeyPair():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_SecKey_Verify(secret_key) == skycoin.SKY_OK


def test_TestGenerateDeterministicKeyPair():
    # deterministic key pairs are useless as is because we can't
    # generate pair n+1, only pair 0
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    _, seed = skycoin.SKY_cipher_RandByte(32)
    skycoin.SKY_cipher_GenerateDeterministicKeyPair(
        seed, public_key, secret_key)
    assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_SecKey_Verify(secret_key) == skycoin.SKY_OK
    skycoin.SKY_cipher_GenerateDeterministicKeyPair(
        seed, public_key, secret_key)
    assert skycoin.SKY_cipher_PubKey_Verify(public_key) == skycoin.SKY_OK
    assert skycoin.SKY_cipher_SecKey_Verify(secret_key) == skycoin.SKY_OK


def test_TestSecKeyTest():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    secret_key_2 = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    value = skycoin.skycoin.SKY_cipher_CheckSecKey(secret_key)
    assert value == skycoin.SKY_OK
    value = skycoin.skycoin.SKY_cipher_CheckSecKey(secret_key_2)
    assert value == skycoin.SKY_ErrInvalidSecKyVerification


def test_TestSecKeyHashTest():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    secret_key_2 = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    sha_sum_1 = skycoin.cipher_SHA256()
    _, data = skycoin.SKY_cipher_RandByte(256)
    skycoin.SKY_cipher_SumSHA256(data, sha_sum_1)
    value = skycoin.skycoin.SKY_cipher_CheckSecKeyHash(secret_key, sha_sum_1)
    assert value == skycoin.SKY_OK
    value = skycoin.skycoin.SKY_cipher_CheckSecKeyHash(secret_key_2, sha_sum_1)
    assert value == skycoin.SKY_ErrInvalidSecKyVerification


def test_TestGenerateDeterministicKeyPairsUsesAllBytes():
    # Tests that if a seed >128 bits is used, the generator does not ignore
    # bits > 128
    seed = b"property diet little foster provide disagree witness mountain alley weekend kitten general"
    secret_keys = skycoin.SKY_cipher_GenerateDeterministicKeyPairsSeed(seed, 3)[
        1:]
    secret_keys_2 = skycoin.SKY_cipher_GenerateDeterministicKeyPairsSeed(seed[:16], 3)[
        1:]
    assert secret_keys != secret_keys_2


def test_TestPubKeyFromHex():
    p1 = skycoin.cipher_PubKey()
    # Invalid hex
    err = skycoin.SKY_cipher_PubKeyFromHex(b"", p1)
    assert err == skycoin.SKY_ErrInvalidLengthPubKey
    err = skycoin.SKY_cipher_PubKeyFromHex(b"cascs", p1)
    assert err == skycoin.SKY_ErrInvalidPubKey


def test_TestPubKeyHex():
    p, sk = utils.makecipher_PubKeyAndcipher_SecKey()
    err, s3 = skycoin.SKY_cipher_PubKey_Hex(p)
    assert err == skycoin.SKY_OK
    p2 = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_PubKeyFromHex(s3, p2)
    assert err == skycoin.SKY_OK
    assert p == p2
    err, s4 = skycoin.SKY_cipher_PubKey_Hex(p2)
    assert err == skycoin.SKY_OK
    assert s3 == s4


def test_TestSecKeyFromHex():
    sk = skycoin.cipher_SecKey()
    # Invalid hex
    err = skycoin.SKY_cipher_SecKeyFromHex(b"", sk)
    assert err == skycoin.SKY_ErrInvalidLengthSecKey
    err = skycoin.SKY_cipher_SecKeyFromHex(b"cascs", sk)
    assert err == skycoin.SKY_ErrInvalidSecKey
    # INvalid hex length
    err, b = skycoin.SKY_cipher_RandByte(32)
    p = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_NewSecKey(b, p)
    assert err == skycoin.SKY_OK



import skycoin
import tests.utils as utils


def test_TestUxBodyHash():
    uxb, _ = utils.makeUxBodyWithSecret()
    hash_null = skycoin.cipher_SHA256()
    hashx = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxBody_Hash(uxb, hashx) == skycoin.SKY_OK
    assert hashx != hash_null


def test_TestUxOutHash():
    uxb, _ = utils.makeUxBodyWithSecret()
    uxo, _ = utils.makeUxOutWithSecret()
    uxo.Body = uxb
    hash_body = skycoin.cipher_SHA256()
    hash_out = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxBody_Hash(uxb, hash_body) == skycoin.SKY_OK
    assert skycoin.SKY_coin_UxOut_Hash(uxo, hash_out) == skycoin.SKY_OK
    assert hash_body == hash_out
    # Head should not affect hash
    uxh = skycoin.coin__UxHead()
    uxh.Time = 0
    uxh.BkSeq = 1
    uxo.Head = uxh
    assert skycoin.SKY_coin_UxOut_Hash(uxo, hash_out) == skycoin.SKY_OK
    assert hash_body == hash_out


def test_TestUxOutSnapshotHash():
    p = skycoin.cipher_PubKey()
    s = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(p, s)
    uxb = skycoin.coin__UxBody()
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_cipher_SumSHA256(b, h) == skycoin.SKY_OK
    uxb.SetSrcTransaction(h.toStr())
    a = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(p, a)
    uxb.Address = a
    uxb.Coins = int(1e6)
    uxb.Hours = int(100)
    uxo = skycoin.coin__UxOut()
    uxh = skycoin.coin__UxHead()
    uxh.Time = 100
    uxh.BkSeq = 2
    uxo.Head = uxh
    uxo.Body = uxb
    hn = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo, hn) == skycoin.SKY_OK
    # snapshot hash should be dependent on every field in body and head
    # Head Time
    uxo_2 = uxo
    uxh.Time = 20
    uxo_2.Head = uxh
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2
    # Head BkSeq
    uxo_2 = uxo
    uxh.BkSeq = 4
    uxo_2.Head = uxh
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2
    # Body SetSrcTransaction
    uxo_2 = uxo
    uxb = skycoin.coin__UxBody()
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_cipher_SumSHA256(b, h) == skycoin.SKY_OK
    uxb.SetSrcTransaction(h.toStr())
    uxo_2.Body = uxb
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2
    # Body Address
    p_2 = skycoin.cipher_PubKey()
    s_2 = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(p_2, s_2)
    a_2 = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(p_2, a_2)
    uxo_2 = uxo
    uxb = skycoin.coin__UxBody()
    uxb.Address = a_2
    uxo_2.Body = uxb
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2
    # Body Coins
    uxo_2 = uxo
    uxb = skycoin.coin__UxBody()
    uxb.Coins = int(2)
    uxo_2.Body = uxb
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2
    # Body Hours
    uxo_2 = uxo
    uxb = skycoin.coin__UxBody()
    uxb.Hours = int(2)
    uxo_2.Body = uxb
    hn_2 = skycoin.cipher_SHA256()
    assert skycoin.SKY_coin_UxOut_SnapshotHash(uxo_2, hn_2) == skycoin.SKY_OK
    assert hn != hn_2


def test_TestUxOutCoinHours():
    p = skycoin.cipher_PubKey()
    s = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(p, s)
    uxb = skycoin.coin__UxBody()
    _, b = skycoin.SKY_cipher_RandByte(128)
    h = skycoin.cipher_SHA256()
    assert skycoin.SKY_cipher_SumSHA256(b, h) == skycoin.SKY_OK
    uxb.SetSrcTransaction(h.toStr())
    a = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(p, a)
    uxb.Address = a
    uxb.Coins = int(1e6)
    uxb.Hours = int(100)
    uxo = skycoin.coin__UxOut()
    uxh = skycoin.coin__UxHead()
    uxh.Time = 100
    uxh.BkSeq = 2
    uxo.Head = uxh
    uxo.Body = uxb

    # Less than an hour passed
    now = uxh.Time + 100
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxh.Time
    assert err == skycoin.SKY_OK
    # 1 hours passed
    now = uxh.Time + 3600
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxh.Time + uxb.Coins / 1000000
    assert err == skycoin.SKY_OK
    # 6 hours passed
    now = uxh.Time + 3600 * 6
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxh.Time + (uxb.Coins / 1000000) * 6
    assert err == skycoin.SKY_OK
    # Time is backwards (treated as no hours passed)
    now = uxh.Time // 2
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxh.Time
    assert err == skycoin.SKY_OK
    # 1 hour has passed, output has 1.5 coins, should gain 1 coinhour
    uxb.Coins = 1500000
    now = uxh.Time + 3600
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours + 1
    assert err == skycoin.SKY_OK
    # 2 hours have passed, output has 1.5 coins, should gain 3 coin hours
    uxb.Coins = 1500000
    uxo.Body = uxb
    now = uxh.Time + 3600 * 2
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours + 3
    assert err == skycoin.SKY_OK
    # 1 second has passed, output has 3600 coins, should gain 1 coin hour
    uxb.Coins = 3600000000
    uxo.Body = uxb
    now = uxh.Time + 1
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours + 1
    assert err == skycoin.SKY_OK
    # 1000000 hours minus 1 second have passed, output has 1 droplet, should
    # gain 0 coin hour
    uxb.Coins = 1
    uxo.Body = uxb
    now = uxh.Time + 1000000 * 3600 - 1
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours
    assert err == skycoin.SKY_OK
    # 1000000 hours have passed, output has 1 droplet, should gain 1 coin hour
    uxb.Coins = 1
    uxo.Body = uxb
    now = uxh.Time + 1000000 * 3600
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours + 1
    assert err == skycoin.SKY_OK
    # No hours passed, using initial coin hours
    uxb.Coins = 1000000000
    uxb.Hours = 1000 * 1000
    uxo.Body = uxb
    now = uxh.Time
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours
    assert err == skycoin.SKY_OK
    # One hour passed, using initial coin hours
    now = uxh.Time + 3600
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == uxb.Hours + 1000000000 / 1000000
    assert err == skycoin.SKY_OK
    # No hours passed and no hours to begin with0
    uxb.Hours = 0
    uxo.Body = uxb
    now = uxh.Time
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert hours == 0
    assert err == skycoin.SKY_OK
    # Centuries have passed, time-based calculation overflows uint64
    # when calculating the whole coin seconds
    uxb.Coins = 2000000
    uxo.Body = uxb
    now = 0xFFFFFFFFFFFFFFFF
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert err == skycoin.SKY_ERROR
    # Centuries have passed, time-based calculation overflows uint64
    # when calculating the droplet seconds
    uxb.Coins = 1500000
    uxo.Body = uxb
    now = 0xFFFFFFFFFFFFFFFF
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert err == skycoin.SKY_ERROR
    # Output would overflow if given more hours, has reached its limit
    uxb.Coins = 3600000000
    uxo.Body = uxb
    now = 0xFFFFFFFFFFFFFFFF
    err, hours = skycoin.SKY_coin_UxOut_CoinHours(uxo, now)
    assert err == skycoin.SKY_ERROR


def test_TestUxArrayCoins():
    uxa = utils.makeUxArray(4)
    for x in uxa:
        x.Body.Coins = utils.Million
    err, coins = skycoin.SKY_coin_UxArray_Coins(uxa)
    assert coins == int(4e6)
    assert err == skycoin.SKY_OK
    uxa[2].Body.Coins = int(utils.MaxUint64 - int(1e6))
    err, _ = skycoin.SKY_coin_UxArray_Coins(uxa)
    assert err == skycoin.SKY_ERROR


def ux_Array_CoinsHours(uxa, now=0, slic=0):
    result = 0
    for x in uxa[slic:]:
        err, time = skycoin.SKY_coin_UxOut_CoinHours(x, now)
        result += time
        assert err == skycoin.SKY_OK
    return result


def test_TestUxArrayCoinHours():
    uxa = utils.makeUxArray(4)
    assert skycoin.SKY_coin_UxArray_CoinHours(uxa, 0)[1] == 400
    # 1 hour later
    assert skycoin.SKY_coin_UxArray_CoinHours(
        uxa, uxa[0].Head.Time + 3600)[1] == 404
    # 1.5 hours later
    assert skycoin.SKY_coin_UxArray_CoinHours(
        uxa, uxa[0].Head.Time + 3600 + 1800)[1] == 404
    # 2 hour later
    assert skycoin.SKY_coin_UxArray_CoinHours(
        uxa, uxa[0].Head.Time + 3600 + 4600)[1] == 408

    uxa[2].Head.Time = utils.MaxUint64 - 100
    # assert skycoin.SKY_coin_UxArray_CoinHours(uxa, utils.MaxUint64 - 100)[1] == skycoin.SKY_OK
    value = skycoin.SKY_coin_UxArray_CoinHours(uxa, uxa[2].Head.Time)[1]
    assert utils.err_CoinHours_Overflow(
        value) == skycoin.SKY_ErrAddEarnedCoinHoursAdditionOverflow
    value = skycoin.SKY_coin_UxArray_CoinHours(uxa, 1000000000000)[1]
    assert utils.err_CoinHours_Overflow(
        value) == skycoin.SKY_ErrAddEarnedCoinHoursAdditionOverflow


def test_TestUxArrayHashArray():
    uxa = utils.makeUxArray(4)
    sha = skycoin.cipher_SHA256()
    err, hashs = skycoin.SKY_coin_UxArray_Hashes(uxa)
    assert err == skycoin.SKY_OK
    assert len(hashs) == len(uxa)
    skycoin.SKY_coin_UxOut_Hash(uxa[0], sha)
    print(sha)
    print(uxa[0])
    assert hashs[0] == sha
    for i in range(len(hashs)):
        assert skycoin.SKY_coin_UxOut_Hash(uxa[i], sha) == 0
        assert sha == hashs[i]


def test_TestUxArrayHasDupes():
    uxa = utils.makeUxArray(4)
    err, hasDupes = skycoin.SKY_coin_UxArray_HasDupes(uxa)
    assert err == skycoin.SKY_OK
    assert hasDupes == 0
    uxa[0] = uxa[1]
    err, hasDupes = skycoin.SKY_coin_UxArray_HasDupes(uxa)
    assert err == skycoin.SKY_OK
    assert hasDupes == 1


def test_TestUxArraySub():
    uxa = utils.makeUxArray(4)
    uxb = utils.makeUxArray(4)
    uxc = uxa[:1]
    for ux in uxb:
        uxc.append(ux)
    for ux in uxa[1:2]:
        uxc.append(ux)
    err, uxd = skycoin.SKY_coin_UxArray_Sub(uxc, uxa)
    assert err == skycoin.SKY_OK
    assert len(uxd) == len(uxb)
    err, uxd = skycoin.SKY_coin_UxArray_Sub(uxc, uxb)
    assert err == skycoin.SKY_OK
    assert len(uxd) == 2
    assert uxd == uxa[:2]
    # No intersection
    err, uxd = skycoin.SKY_coin_UxArray_Sub(uxa, uxb)
    assert uxd == uxa
    err, uxd = skycoin.SKY_coin_UxArray_Sub(uxb, uxa)
    assert uxd == uxb


def manualUxArrayIsSorted(uxa):
    sha_1 = skycoin.cipher_SHA256()
    sha_2 = skycoin.cipher_SHA256()
    isSorte = True
    for i in range(len(uxa) - 1):
        assert skycoin.SKY_coin_UxOut_Hash(uxa[i], sha_1) == skycoin.SKY_OK
        assert skycoin.SKY_coin_UxOut_Hash(uxa[i + 1], sha_2) == skycoin.SKY_OK
        if sha_1 > sha_2:
            isSorte = False
    return isSorte


def isUxArraySorted(uxa):
    n = len(uxa)
    prev = uxa
    current = prev
    current += 1
    hash_1 = skycoin.cipher_SHA256()
    hash_2 = skycoin.cipher_SHA256()
    prevHash = None
    currentHash = None
    result = int()
    for i in n:
        if(prevHash is None):
            result = skycoin.SKY_coin_UxOut_Hash(prev, hash_1)
            assert result == skycoin.SKY_OK
            prevHash = hash_1
        if currentHash is None:
            currentHash = hash_2
            result = skycoin.SKY_coin_UxOut_Hash(current, currentHash)
            assert result == skycoin.SKY_OK
        if prevHash.__eq__(currentHash) > 0:
            return 0
        if i % 2 != 0:
            prevHash = hash_2
            currentHash = hash_1
        else:
            prevHash = hash_1
            currentHash = hash_2
        prev += 1
        current += 1
    return 1


def test_TestUxArrayLen():
    uxa = utils.makeUxArray(4)
    assert len(uxa) == 4


def test_TestUxArrayLess():
    uxa = utils.makeUxArray(2)
    err, hasha = skycoin.SKY_coin_UxArray_Hashes(uxa)
    assert err == skycoin.SKY_OK and len(hasha) == len(uxa)
    err, _ = skycoin.SKY_coin_UxArray_Less(uxa, 0, 1)
    assert err == skycoin.SKY_OK
    err, _ = skycoin.SKY_coin_UxArray_Less(uxa, 1, 0)
    assert err == skycoin.SKY_OK


def test_TestUxArraySwap():
    uxa = utils.makeUxArray(2)
    uxx = utils.make_UxOut()
    uxy = utils.make_UxOut()
    uxa[0] = uxx
    uxa[1] = uxy
    err = skycoin.SKY_coin_UxArray_Swap(uxa, 0, 1)
    assert err == skycoin.SKY_OK
    uxa[0] = uxy
    uxa[1] = uxx
    err = skycoin.SKY_coin_UxArray_Swap(uxa, 0, 1)
    assert err == skycoin.SKY_OK
    uxa[0] = uxx
    uxa[1] = uxy
    err = skycoin.SKY_coin_UxArray_Swap(uxa, 0, 1)
    assert err == skycoin.SKY_OK
    uxa[1] = uxx
    uxa[0] = uxy


def test_TestAddressUxOutsKeys():
    uxa = utils.makeUxArray(3)
    err, uxH = skycoin.SKY_coin_NewAddressUxOuts(uxa)
    assert err == skycoin.SKY_OK
    # keys = []
    err, keys = skycoin.SKY_coin_AddressUxOuts_Keys(uxH)
    assert err == skycoin.SKY_OK
    assert len(keys) == 3
    for k in keys:
        assert k == uxa[0].Body.Address or k == uxa[1].Body.Address or k == uxa[2].Body.Address


def test_TestAddressUxOutsSub():
    uxa = utils.makeUxArray(4)
    empty = utils.makeUxArray(0)
    err, uxH_1 = skycoin.SKY_coin_NewAddressUxOuts(empty)
    assert err == skycoin.SKY_OK
    err, uxH_2 = skycoin.SKY_coin_NewAddressUxOuts(empty)
    assert err == skycoin.SKY_OK
    uxa[1].Body.Address = uxa[0].Body.Address
    ux_2 = uxa[:2]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[0].Body.Address, ux_2)
    assert err == skycoin.SKY_OK
    ux_3 = [uxa[2]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[2].Body.Address, ux_3)
    assert err == skycoin.SKY_OK
    ux_4 = [uxa[3]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[3].Body.Address, ux_4)
    assert err == skycoin.SKY_OK

    ux_5 = [uxa[0]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_2, uxa[0].Body.Address, ux_5)
    assert err == skycoin.SKY_OK
    ux_6 = [uxa[2]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_2, uxa[2].Body.Address, ux_6)
    assert err == skycoin.SKY_OK
    err, uxH_3 = skycoin.SKY_coin_AddressUxOuts_Sub(uxH_1, uxH_2)
    assert err == skycoin.SKY_OK
    # length
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_3)
    assert length == 2
    assert err == skycoin.SKY_OK
    # hasKey
    err, has_key = skycoin.SKY_coin_AddressUxOuts_HasKey(
        uxH_3, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert has_key == 0
    err, ux_3 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[3].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_3) == 1
    uxtmp = ux_3[0]
    assert uxtmp == ux_3[0]
    assert ux_3[0] == uxa[3]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 1
    assert ux_2[0] == uxa[1]
    # Originals should be unmodified
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_1)
    assert err == skycoin.SKY_OK
    assert length == 3
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 2
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[3].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1

    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_2)
    assert err == skycoin.SKY_OK
    assert length == 2
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_2, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_2, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1


def test_TestAddressUxOutsAdd():
    uxa = utils.makeUxArray(4)
    empty = utils.makeUxArray(0)
    err, uxH_1 = skycoin.SKY_coin_NewAddressUxOuts(empty)
    assert err == skycoin.SKY_OK
    err, uxH_2 = skycoin.SKY_coin_NewAddressUxOuts(empty)
    assert err == skycoin.SKY_OK
    uxa[1].Body.Address = uxa[0].Body.Address
    ux_2 = [uxa[0]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[0].Body.Address, ux_2)
    assert err == skycoin.SKY_OK
    ux_3 = [uxa[2]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[2].Body.Address, ux_3)
    assert err == skycoin.SKY_OK
    ux_4 = [uxa[3]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_1, uxa[3].Body.Address, ux_4)
    assert err == skycoin.SKY_OK
    ux_5 = [uxa[0]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_2, uxa[0].Body.Address, ux_5)
    assert err == skycoin.SKY_OK
    ux_6 = [uxa[2]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH_2, uxa[2].Body.Address, ux_6)
    assert err == skycoin.SKY_OK
    err, uxH_3 = skycoin.SKY_coin_AddressUxOuts_Add(uxH_1, uxH_2)
    assert err == skycoin.SKY_OK
    # length
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_3)
    assert err == skycoin.SKY_OK
    assert length == 3
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_3, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    # assert length == 2
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    # assert len(ux_2) == 2
    assert ux_2[0] == uxa[0]
    # assert ux_2[1] == uxa[1]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 1
    assert ux_2[0] == uxa[2]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[3].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 1
    assert ux_2[0] == uxa[3]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH_3, uxa[1].Body.Address)
    assert err == skycoin.SKY_OK
    # assert len(ux_2) == 2
    assert ux_2[0] == uxa[0]
    # assert ux_2[1] == uxa[1]
    # Originals should be unmodified
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_1)
    assert err == skycoin.SKY_OK
    assert length == 3
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_1, uxa[3].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH_2)
    assert err == skycoin.SKY_OK
    assert length == 2
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_2, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1
    err, length = skycoin.SKY_coin_AddressUxOuts_GetOutputLength(
        uxH_2, uxa[2].Body.Address)
    assert err == skycoin.SKY_OK
    assert length == 1


def test_TestAddressUxOutsFlatten():
    uxa = utils.makeUxArray(3)
    empty = utils.makeUxArray(0)
    err, uxH = skycoin.SKY_coin_NewAddressUxOuts(empty)
    assert err == skycoin.SKY_OK
    uxa[2].Body.Address = uxa[1].Body.Address
    emptyAddr = skycoin.cipher__Address()
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH, emptyAddr, empty)
    assert err == skycoin.SKY_OK
    ux_1 = [uxa[0]]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH, uxa[0].Body.Address, ux_1)
    assert err == skycoin.SKY_OK
    ux_2 = uxa[1:]
    err = skycoin.SKY_coin_AddressUxOuts_Set(uxH, uxa[1].Body.Address, ux_2)
    assert err == skycoin.SKY_OK
    err, flatArray = skycoin.SKY_coin_AddressUxOuts_Flatten(uxH)
    assert err == skycoin.SKY_OK
    assert len(flatArray) == 3
    for x in flatArray:
        assert x.Body.Address != emptyAddr

    if flatArray[0].Body.Address == uxa[0].Body.Address:
        assert flatArray[0] == uxa[0]
        assert flatArray[0].Body.Address == uxa[0].Body.Address
        assert flatArray[1] == uxa[1]
        assert flatArray[1].Body.Address == uxa[1].Body.Address
        assert flatArray[2] == uxa[2]
        assert flatArray[2].Body.Address == uxa[2].Body.Address
    else:
        assert flatArray[0] == uxa[1]
        assert flatArray[0].Body.Address == uxa[1].Body.Address
        assert flatArray[1] == uxa[2]
        assert flatArray[1].Body.Address == uxa[2].Body.Address
        assert flatArray[2] == uxa[0]
        assert flatArray[2].Body.Address == uxa[0].Body.Address


def test_TestNewAddressUxOuts():
    uxa = utils.makeUxArray(6)
    uxa[1].Body.Address = uxa[0].Body.Address
    uxa[3].Body.Address = uxa[2].Body.Address
    uxa[4].Body.Address = uxa[2].Body.Address
    err, uxH = skycoin.SKY_coin_NewAddressUxOuts(uxa)
    assert err == skycoin.SKY_OK
    # length
    err, length = skycoin.SKY_coin_AddressUxOuts_Length(uxH)
    assert err == skycoin.SKY_OK
    assert length == 3
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH, uxa[0].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 2
    assert ux_2[0] == uxa[0]
    assert ux_2[1] == uxa[1]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH, uxa[3].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 3
    assert ux_2[0] == uxa[2]
    assert ux_2[1] == uxa[3]
    assert ux_2[2] == uxa[4]
    err, ux_2 = skycoin.SKY_coin_AddressUxOuts_Get(uxH, uxa[5].Body.Address)
    assert err == skycoin.SKY_OK
    assert len(ux_2) == 1
    assert ux_2[0] == uxa[5]



__version__ = "0.26.0.dev1"
init_error = None

def _print2stderr(msg):
    sys.stderr.write(msg + '\n')

try:
    from .skycoin import *
except (AttributeError, ImportError) as _err :
    init_error = _err

    import sys, traceback
    _print2stderr("Error initializing skycoin package. Details:")
    traceback.print_exception(*sys.exc_info())
    _print2stderr("\n\nInport succeeded but package load failed")




import skycoin
import tests.utils as utils


def test_TestBitcoinAddressFromBytes():
    p, s = utils.makecipher_PubKeyAndcipher_SecKey()
    a = skycoin.cipher__BitcoinAddress()
    err = skycoin.SKY_cipher_BitcoinAddressFromSecKey(s, a)
    assert err == skycoin.SKY_OK
    # Valid pubkey+address
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(a, p)
    assert err == skycoin.SKY_OK
    s = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_BitcoinAddressFromSecKey(s, a)
    assert err == skycoin.SKY_ErrPubKeyFromNullSecKey


def test_TestBitcoinWIPRoundTrio():
    public_key = skycoin.cipher_PubKey()
    secret_key_1 = skycoin.cipher_SecKey()
    secret_key_2 = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key_1)
    wip_1 = skycoin.skycoin.SKY_cipher_BitcoinWalletImportFormatFromSeckey(
        secret_key_1)
    err = skycoin.skycoin.SKY_cipher_SecKeyFromBitcoinWalletImportFormat(
        wip_1, secret_key_2)
    assert err == skycoin.SKY_OK
    wip_2 = skycoin.skycoin.SKY_cipher_BitcoinWalletImportFormatFromSeckey(
        secret_key_2)
    _, secret_key_1_hex = skycoin.SKY_cipher_SecKey_Hex(secret_key_1)
    _, secret_key_2_hex = skycoin.SKY_cipher_SecKey_Hex(secret_key_2)
    assert secret_key_1_hex == secret_key_2_hex
    assert wip_1 == wip_2


def test_TestBitcoinWIF():
    wips = [
        b"KwntMbt59tTsj8xqpqYqRRWufyjGunvhSyeMo3NTYpFYzZbXJ5Hp",
        b"L4ezQvyC6QoBhxB4GVs9fAPhUKtbaXYUn8YTqoeXwbevQq4U92vN",
        b"KydbzBtk6uc7M6dXwEgTEH2sphZxSPbmDSz6kUUHi4eUpSQuhEbq"
    ]

    publics = [
        b"034f355bdcb7cc0af728ef3cceb9615d90684bb5b2ca5f859ab0f0b704075871aa",
        b"02ed83704c95d829046f1ac27806211132102c34e9ac7ffa1b71110658e5b9d1bd",
        b"032596957532fc37e40486b910802ff45eeaa924548c0e1c080ef804e523ec3ed3"
    ]

    address = [
        b"1Q1pE5vPGEEMqRcVRMbtBK842Y6Pzo6nK9",
        b"1NKRhS7iYUGTaAfaR5z8BueAJesqaTyc4a",
        b"19ck9VKC6KjGxR9LJg4DNMRc45qFrJguvV"
    ]

    for p in range(len(wips)):
        secret_key = skycoin.cipher_SecKey()
        public_key = skycoin.cipher_PubKey()
        err = skycoin.skycoin.SKY_cipher_SecKeyFromBitcoinWalletImportFormat(
            wips[p], secret_key)
        assert err == skycoin.SKY_OK
        skycoin.SKY_cipher_PubKeyFromSecKey(secret_key, public_key)
        _, public_key_hex = skycoin.SKY_cipher_PubKey_Hex(public_key)
        assert public_key_hex == publics[p]
        bitcoin_addr = skycoin.cipher__BitcoinAddress()
        skycoin.SKY_cipher_BitcoinAddressFromPubKey(public_key, bitcoin_addr)
        bitcoin_addr_str = skycoin.skycoin.SKY_cipher_BitcoinAddress_String(
            bitcoin_addr)
        assert bitcoin_addr_str == address[p]


def test_TestDecodeBase58BitcoinAddress():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__BitcoinAddress()
    skycoin.SKY_cipher_BitcoinAddressFromPubKey(public_key, address)
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(address, public_key)
    assert err == skycoin.SKY_OK
    address_2 = skycoin.cipher__BitcoinAddress()
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(b'""', address_2)
    assert err == skycoin.SKY_ERROR
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(b'"cascs"', address_2)
    assert err == skycoin.SKY_ERROR
    byte = skycoin.SKY_cipher_BitcoinAddress_Bytes(address)
    _, h = skycoin.SKY_base58_Hex2Base58(byte[:int(len(byte) / 2)])
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(h, address_2)
    assert err == skycoin.SKY_ErrAddressInvalidLength
    _, h = skycoin.SKY_base58_Hex2Base58(byte)
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(h, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    addres_str = skycoin.SKY_cipher_BitcoinAddress_String(address)
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(addres_str, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    #  preceding whitespace is invalid
    addres_2_str = b'" " + a_str'
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(
        addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    #  preceding zeroes are invalid
    addres_2_str = b'"000" + a_str'
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(
        addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    #  trailing whitespace is invalid
    addres_2_str = b'a_str + " "'
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(
        addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    # trailing zeroes are invalid
    addres_2_str = b'a_str + "000"'
    err = skycoin.SKY_cipher_DecodeBase58BitcoinAddress(
        addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR


def test_TestBitcoinAddressNull():
    a = skycoin.cipher__BitcoinAddress()
    err = skycoin.SKY_cipher_BitcoinAddress_Null(a)
    assert err == 1
    p = skycoin.cipher_PubKey()
    s = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_GenerateKeyPair(p, s)
    assert err == skycoin.SKY_OK
    skycoin.SKY_cipher_BitcoinAddressFromPubKey(p, a)
    err = skycoin.SKY_cipher_BitcoinAddress_Null(a)
    assert err == 0


def test_TestBitcoinAddressVerify():
    p = skycoin.cipher_PubKey()
    s = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_GenerateKeyPair(p, s)
    assert err == skycoin.SKY_OK
    a = skycoin.cipher__BitcoinAddress()
    skycoin.SKY_cipher_BitcoinAddressFromPubKey(p, a)
    # Valid pubkey+address
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(a, p)
    assert err == skycoin.SKY_OK
    # Invalid pubkey
    p = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(a, p)
    assert err == skycoin.SKY_ErrAddressInvalidPubKey
    p2 = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_GenerateKeyPair(p2, s)
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(a, p2)
    assert err == skycoin.SKY_ErrAddressInvalidPubKey
    # Bad version
    a.Version = 0x01
    err = skycoin.SKY_cipher_BitcoinAddress_Verify(a, p2)
    assert err == skycoin.SKY_ErrAddressInvalidVersion


def test_TestBitcoinWIFRoundTrip():
    p, s = utils.makecipher_PubKeyAndcipher_SecKey()
    wip1 = skycoin.SKY_cipher_BitcoinWalletImportFormatFromSeckey(s)
    s2 = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_SecKeyFromBitcoinWalletImportFormat(wip1, s2)
    assert err == skycoin.SKY_OK
    wip2 = skycoin.SKY_cipher_BitcoinWalletImportFormatFromSeckey(s2)
    assert s == s2
    sh = skycoin.SKY_cipher_SecKey_Hex(s)
    sh2 = skycoin.SKY_cipher_SecKey_Hex(s2)
    assert sh == sh2
    assert wip1 == wip2



import skycoin
import json
import base64
import tests.utils

define = {
    "PLAINTEXT": b"plaintext",
    "PASSWORD": b"password",
    "ENCRYPTED": b"dQB7Im4iOjUyNDI4OCwiciI6OCwicCI6MSwia2V5TGVuIjozMiwic2FsdCI6ImpiejUrSFNjTFFLWkI5T0tYblNNRmt2WDBPY3JxVGZ0ZFpDNm9KUFpaeHc9Iiwibm9uY2UiOiJLTlhOQmRQa1ZUWHZYNHdoIn3PQFmOot0ETxTuv//skTG7Q57UVamGCgG5",
    "SCRYPTCHACHA20METALENGTHSIZE": 2,
}


def test_TestScryptChacha20poly1305Encrypt():
    for i in range(20)[1:]:
        crypto = skycoin.encrypt__ScryptChacha20poly1305()
        crypto.R = 8
        crypto.P = 1
        crypto.KeyLen = 32
        crypto.N = 1 << i
        _, encData = skycoin.SKY_encrypt_ScryptChacha20poly1305_Encrypt(
            crypto, define["PLAINTEXT"], define["PASSWORD"])
        assert _ == skycoin.SKY_OK
        Data = base64.standard_b64decode(encData)
        ml_ = [x for x in Data]
        if isinstance(ml_[0], int):
            ml = ml_[0]
            _Data = [chr(x) for x in Data]
            D = ""
            for x in _Data:
                D += x
            Data = D
        else:
            ml = ord(ml_[0])
        assert int(define["SCRYPTCHACHA20METALENGTHSIZE"] + ml) <= len(Data)
        m = json.loads(Data[define["SCRYPTCHACHA20METALENGTHSIZE"]
                       :define["SCRYPTCHACHA20METALENGTHSIZE"] + ml])
        print(m)
        assert m["n"] == 1 << i
        assert m["r"] == 8
        assert m["p"] == 1
        assert m["keyLen"] == 32


def test_TestScryptChacha20poly1305Decrypt():
    encrypto = b"dQB7Im4iOjUyNDI4OCwiciI6OCwicCI6MSwia2V5TGVuIjozMiwic2FsdCI6ImpiejUrSFNjTFFLWkI5T0tYblNNRmt2WDBPY3JxVGZ0ZFpDNm9KUFpaeHc9Iiwibm9uY2UiOiJLTlhOQmRQa1ZUWHZYNHdoIn3PQFmOot0ETxTuv//skTG7Q57UVamGCgG5"
    password = b"pwd"
    invalid_passwd = b"wrong password"
    crypto = skycoin.encrypt__ScryptChacha20poly1305()
    crypto.R = 8
    crypto.P = 1
    crypto.KeyLen = 32
    crypto.N = 1 << 19
    err, decrypted = skycoin.SKY_encrypt_ScryptChacha20poly1305_Decrypt(
        crypto, encrypto, password)
    assert err == skycoin.SKY_OK and decrypted == define["PLAINTEXT"]
    # Wrong Password
    err, decrypted = skycoin.SKY_encrypt_ScryptChacha20poly1305_Decrypt(
        crypto, encrypto, invalid_passwd)
    assert err == skycoin.SKY_ERROR
    # Missing Password
    err, decrypted = skycoin.SKY_encrypt_ScryptChacha20poly1305_Decrypt(
        crypto, encrypto, b"")
    assert err == skycoin.SKY_ERROR



# import skycoin
# from . import test_testsuite
# import json
# import os
# import re

# # testdataDir           = "./testdata/"
# testdataDir           = "gopath/src/github.com/skycoin/libskycoin/vendor/github.com/skycoin/skycoin/src/cipher/testsuite/testdata/"
# manyAddressesFilename = "many-addresses.golden"
# inputHashesFilename   = "input-hashes.golden"
# seedFileRegex         = "seed-\d+.golden"


# def test_TestManyAddresses():
#     f = open(testdataDir+manyAddressesFilename,'r')
#     fn = f.read()
#     f_JSON = json.loads(fn)
#     dataJSON = test_testsuite.SeedTestData()
#     dataJSON.Seed = f_JSON["seed"]
#     dataJSON.Keys = f_JSON["keys"]
#     err, data = test_testsuite.SeedTestDataFromJSON(dataJSON)
#     assert err == skycoin.SKY_OK
#     err = test_testsuite.ValidateSeedData(data, None)
#     assert err == skycoin.SKY_OK

# def test_TestSeedSignatures():
#     f = open(testdataDir+inputHashesFilename,'r')
#     fn = f.read()
#     f_JSON = json.loads(fn)
#     inputDataJSON = test_testsuite.InputTestData()
#     inputDataJSON.Hashes = f_JSON["hashes"]
#     err, inputData = test_testsuite.InputTestDataFromJSON(inputDataJSON)
#     assert err == skycoin.SKY_OK
#     err, seedFiles = traverseFiles()
#     assert err == skycoin.SKY_OK
#     for dir_f in seedFiles:
#         f = open(testdataDir+dir_f,'r')
#         fn = f.read()
#         f_JSON = json.loads(fn)
#         dataJSON = test_testsuite.SeedTestData()
#         dataJSON.Seed = f_JSON["seed"]
#         dataJSON.Keys = f_JSON["keys"]
#         err, data = test_testsuite.SeedTestDataFromJSON(dataJSON)
#         assert err == skycoin.SKY_OK
#         err = test_testsuite.ValidateSeedData(data, inputData)
#         assert err == skycoin.SKY_OK


# def traverseFiles():
#     files = []
#     root = os.listdir(testdataDir)
#     pattern = re.compile(seedFileRegex)
#     for path in root:
#         if pattern.match(path):
#             files.append(path)
#     if len(files) == 0:
#         return skycoin.SKY_ERROR, None
#     return skycoin.SKY_OK, files



import skycoin


def test_TestDecodeBase58Address():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    err = skycoin.SKY_cipher_Address_Verify(address, public_key)
    assert err == skycoin.SKY_OK
    address_2 = skycoin.cipher__Address()
    err = skycoin.SKY_cipher_DecodeBase58Address(b'""', address_2)
    assert err == skycoin.SKY_ERROR
    err = skycoin.SKY_cipher_DecodeBase58Address(b'"cascs"', address_2)
    assert err == skycoin.SKY_ERROR
    _, byte = skycoin.SKY_cipher_Address_Bytes(address)
    _, h = skycoin.SKY_base58_Hex2Base58(byte[:int(len(byte) / 2)])
    err = skycoin.SKY_cipher_DecodeBase58Address(h, address_2)
    assert err == skycoin.SKY_ErrAddressInvalidLength
    _, h = skycoin.SKY_base58_Hex2Base58(byte)
    err = skycoin.SKY_cipher_DecodeBase58Address(h, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    _, addres_str = skycoin.SKY_cipher_Address_String(address)
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_str, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    #  preceding whitespace is invalid
    addres_2_str = b'" " + a_str'
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    #  preceding zeroes are invalid
    addres_2_str = b'"000" + a_str'
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    #  trailing whitespace is invalid
    addres_2_str = b'a_str + " "'
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR
    # trailing zeroes are invalid
    addres_2_str = b'a_str + "000"'
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_2_str, address_2)
    assert err == skycoin.SKY_ERROR


def test_TestAddressFromBytes():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    address_2 = skycoin.cipher__Address()
    _, byte = skycoin.SKY_cipher_Address_Bytes(address)
    err = skycoin.SKY_cipher_AddressFromBytes(byte, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    # Invalid number of bytes
    __ = skycoin.cipher__Address()
    err = skycoin.SKY_cipher_AddressFromBytes(byte[:len(byte) - 2], __)
    assert err == skycoin.SKY_ErrAddressInvalidLength
    # Invalid checksum
    byte_array = bytearray(byte)
    byte_array[-1] = 1
    byte_new = bytes(byte_array)
    err = skycoin.SKY_cipher_AddressFromBytes(byte_new, __)
    assert err == skycoin.SKY_ErrAddressInvalidChecksum
    address.Version = 2
    _, b = skycoin.SKY_cipher_Address_Bytes(address)
    err = skycoin.SKY_cipher_AddressFromBytes(b, __)
    assert err == skycoin.SKY_ErrAddressInvalidVersion


def test_TestAddressRoundtrip():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__Address()
    address_2 = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    _, byte = skycoin.skycoin.SKY_cipher_Address_Bytes(address)
    err = skycoin.skycoin.SKY_cipher_AddressFromBytes(byte, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    _, addres_str = skycoin.SKY_cipher_Address_String(address)
    _, addres_2_str = skycoin.SKY_cipher_Address_String(address_2)
    assert addres_2_str == addres_str


def test_TestAddressVerify():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    # Valid pubkey+address
    err = skycoin.SKY_cipher_Address_Verify(address, public_key)
    assert err == skycoin.SKY_OK
    # Invalid pubkey
    public_key_temp = skycoin.cipher_PubKey()
    err = skycoin.SKY_cipher_Address_Verify(address, public_key_temp)
    assert err == skycoin.SKY_ErrAddressInvalidPubKey
    skycoin.SKY_cipher_GenerateKeyPair(public_key_temp, secret_key)
    err = skycoin.SKY_cipher_Address_Verify(address, public_key_temp)
    assert err == skycoin.SKY_ErrAddressInvalidPubKey
    #  Bad version
    address.Version = 0x01
    err = skycoin.SKY_cipher_Address_Verify(address, public_key)
    assert err == skycoin.SKY_ErrAddressInvalidVersion


def test_TestAddressString():
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    address = skycoin.cipher__Address()
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    _, addres_str = skycoin.SKY_cipher_Address_String(address)
    address_2 = skycoin.cipher__Address()
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_str, address_2)
    assert err == skycoin.SKY_OK
    assert address == address_2
    _, addres_2_str = skycoin.SKY_cipher_Address_String(address_2)
    addres_3 = skycoin.cipher__Address()
    err = skycoin.SKY_cipher_DecodeBase58Address(addres_2_str, addres_3)
    assert err == skycoin.SKY_OK
    assert address_2 == addres_3


def test_TestBitcoinAddress2():
    secret_key = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_SecKeyFromHex(
        b'dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd', secret_key)
    assert err == skycoin.SKY_OK
    public_key = skycoin.cipher_PubKey()
    skycoin.SKY_cipher_PubKeyFromSecKey(secret_key, public_key)
    public_key_str = b'02ed83704c95d829046f1ac27806211132102c34e9ac7ffa1b71110658e5b9d1bd'
    _, public_key_hex = skycoin.SKY_cipher_PubKey_Hex(public_key)
    assert public_key_hex == public_key_str
    bitcoin_srt = b'1NKRhS7iYUGTaAfaR5z8BueAJesqaTyc4a'
    bitcoin_addr = skycoin.cipher__BitcoinAddress()
    skycoin.SKY_cipher_BitcoinAddressFromPubKey(public_key, bitcoin_addr)
    bitcoin_addr_str = skycoin.skycoin.SKY_cipher_BitcoinAddress_String(
        bitcoin_addr)
    assert bitcoin_srt == bitcoin_addr_str


def test_TestBitcoinAddress3():
    secret_key = skycoin.cipher_SecKey()
    err = skycoin.SKY_cipher_SecKeyFromHex(
        b'47f7616ea6f9b923076625b4488115de1ef1187f760e65f89eb6f4f7ff04b012', secret_key)
    assert err == skycoin.SKY_OK
    public_key = skycoin.cipher_PubKey()
    skycoin.SKY_cipher_PubKeyFromSecKey(secret_key, public_key)
    public_key_str = b'032596957532fc37e40486b910802ff45eeaa924548c0e1c080ef804e523ec3ed3'
    _, public_key_hex = skycoin.SKY_cipher_PubKey_Hex(public_key)
    assert public_key_hex == public_key_str
    bitcoin_srt = b'19ck9VKC6KjGxR9LJg4DNMRc45qFrJguvV'
    bitcoin_addr = skycoin.cipher__BitcoinAddress()
    skycoin.SKY_cipher_BitcoinAddressFromPubKey(public_key, bitcoin_addr)
    bitcoin_addr_str = skycoin.skycoin.SKY_cipher_BitcoinAddress_String(
        bitcoin_addr)
    assert bitcoin_srt == bitcoin_addr_str


def test_TestAddressBulk():
    for _ in range(1024):
        public_key = skycoin.cipher_PubKey()
        secret_key = skycoin.cipher_SecKey()
        addres_1 = skycoin.cipher__Address()
        address_2 = skycoin.cipher__Address()
        _, data = skycoin.SKY_cipher_RandByte(32)
        skycoin.SKY_cipher_GenerateDeterministicKeyPair(
            data, public_key, secret_key)
        skycoin.SKY_cipher_AddressFromPubKey(public_key, addres_1)
        err = skycoin.SKY_cipher_Address_Verify(addres_1, public_key)
        assert err == skycoin.SKY_OK
        _, addres_str = skycoin.SKY_cipher_Address_String(addres_1)
        err = skycoin.SKY_cipher_DecodeBase58Address(addres_str, address_2)
        assert err == skycoin.SKY_OK
        assert addres_1 == address_2


def test_TestAddressNull():
    address = skycoin.cipher__Address()
    _error, isNull = skycoin.SKY_cipher_Address_Null(address)
    assert _error == skycoin.SKY_OK
    assert isNull == 1
    public_key = skycoin.cipher_PubKey()
    secret_key = skycoin.cipher_SecKey()
    skycoin.SKY_cipher_GenerateKeyPair(public_key, secret_key)
    skycoin.SKY_cipher_AddressFromPubKey(public_key, address)
    assert address is not None
